# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging

from bentoml.adapters import (
    BaseInputAdapter,
    ClipperBytesInput,
    ClipperDoublesInput,
    ClipperFloatsInput,
    ClipperIntsInput,
    ClipperStringsInput,
    DataframeInput,
    JsonInput,
    MultiImageInput,
    TfTensorInput,
)

logger = logging.getLogger(__name__)

logger.warning(
    'bentoml.handlers.* will be deprecated after BentoML 1.0, '
    'use bentoml.adapters.* instead'
)


def deprecated(cls, cls_name):
    class wrapped_cls(cls):
        def __init__(self, *args, **kwargs):
            super(wrapped_cls, self).__init__(*args, **kwargs)
            logger.warning(
                f'{cls_name} will be deprecated after BentoML 1.0, '
                f'use {cls.__name__} instead'
            )

    for attr_name in functools.WRAPPER_ASSIGNMENTS:
        setattr(wrapped_cls, attr_name, getattr(cls, attr_name, None))
    return wrapped_cls


BentoHandler = deprecated(BaseInputAdapter, 'BentoHandler')
DataframeHandler = deprecated(DataframeInput, 'DataframeHandler')
TensorflowTensorHandler = deprecated(TfTensorInput, 'TensorflowTensorHandler')
JsonHandler = deprecated(JsonInput, 'JsonHandler')
ImageHandler = deprecated(MultiImageInput, 'ImageHandler')
ClipperIntsHandler = deprecated(ClipperIntsInput, 'ClipperIntsHandler')
ClipperBytesHandler = deprecated(ClipperBytesInput, 'ClipperBytesHandler')
ClipperDoublesHandler = deprecated(ClipperDoublesInput, 'ClipperDoublesHandler')
ClipperFloatsHandler = deprecated(ClipperFloatsInput, 'ClipperFloatsHandler')
ClipperStringsHandler = deprecated(ClipperStringsInput, 'ClipperStringsHandler')

__all__ = [
    "BentoHandler",
    "DataframeHandler",
    "TensorflowTensorHandler",
    "JsonHandler",
    "ImageHandler",
    "ClipperBytesHandler",
    "ClipperDoublesHandler",
    "ClipperFloatsHandler",
    "ClipperIntsHandler",
    "ClipperStringsHandler",
]



DEFAULT_MIN_SIZE = 1
DEFAULT_DESIRED_CAPACITY = 1
DEFAULT_MAX_SIZE = 1

DEFAULT_INSTANCE_TYPE = "t2.micro"
DEFAULT_AMI_ID = "/aws/service/ami-amazon-linux-latest/amzn2-ami-hvm-x86_64-gp2"

BENTOSERVICE_PORT = 5000

AWS_EC2_IN_SERVICE_STATE = "InService"

TARGET_HEALTHY_STATUS = "healthy"
TARGET_UNHEALTHY_STATUS = "unhealthy"
TARGET_INITIAL_STATUS = "initial"
TARGET_UNUSED_STATUS = "unused"
TARGET_UNAVAILABLE_STATUS = "unavailable"
TARGET_DRAINIGN_STATUS = "draining"

TARGET_HEALTH_CHECK_INTERVAL = 5
TARGET_HEALTH_CHECK_PATH = "/healthz"
TARGET_HEALTH_CHECK_PORT = 5000
TARGET_HEALTH_CHECK_TIMEOUT_SECONDS = 3
TARGET_HEALTH_CHECK_THRESHOLD_COUNT = 2



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import sys
import tarfile
import logging
import tempfile
import shutil
from functools import wraps
from contextlib import contextmanager
from urllib.parse import urlparse
from typing import TYPE_CHECKING
from pathlib import PureWindowsPath, PurePosixPath

from bentoml.utils.s3 import is_s3_url
from bentoml.utils.gcs import is_gcs_url
from bentoml.exceptions import BentoMLException
from bentoml.saved_bundle.config import SavedBundleConfig
from bentoml.saved_bundle.pip_pkg import ZIPIMPORT_DIR

if TYPE_CHECKING:
    from bentoml.yatai.proto.repository_pb2 import BentoServiceMetadata

logger = logging.getLogger(__name__)


def _is_http_url(bundle_path) -> bool:
    try:
        return urlparse(bundle_path).scheme in ["http", "https"]
    except ValueError:
        return False


def _is_remote_path(bundle_path) -> bool:
    return isinstance(bundle_path, str) and (
        is_s3_url(bundle_path) or is_gcs_url(bundle_path) or _is_http_url(bundle_path)
    )


@contextmanager
def _resolve_remote_bundle_path(bundle_path):
    if is_s3_url(bundle_path):
        import boto3

        parsed_url = urlparse(bundle_path)
        bucket_name = parsed_url.netloc
        object_name = parsed_url.path.lstrip('/')

        s3 = boto3.client('s3')
        fileobj = io.BytesIO()
        s3.download_fileobj(bucket_name, object_name, fileobj)
        fileobj.seek(0, 0)
    elif is_gcs_url(bundle_path):
        try:
            from google.cloud import storage
        except ImportError:
            raise BentoMLException(
                '"google-cloud-storage" package is required. You can install it with '
                'pip: "pip install google-cloud-storage"'
            )

        gcs = storage.Client()
        fileobj = io.BytesIO()
        gcs.download_blob_to_file(bundle_path, fileobj)
        fileobj.seek(0, 0)
    elif _is_http_url(bundle_path):
        import requests

        response = requests.get(bundle_path)
        if response.status_code != 200:
            raise BentoMLException(
                f"Error retrieving BentoService bundle. "
                f"{response.status_code}: {response.text}"
            )
        fileobj = io.BytesIO()
        fileobj.write(response.content)
        fileobj.seek(0, 0)
    else:
        raise BentoMLException(f"Saved bundle path: '{bundle_path}' is not supported")

    with tarfile.open(mode="r:gz", fileobj=fileobj) as tar:
        with tempfile.TemporaryDirectory() as tmpdir:
            filename = tar.getmembers()[0].name
            tar.extractall(path=tmpdir)
            yield os.path.join(tmpdir, filename)


def resolve_remote_bundle(func):
    """Decorate a function to handle remote bundles."""

    @wraps(func)
    def wrapper(bundle_path, *args):
        if _is_remote_path(bundle_path):
            with _resolve_remote_bundle_path(bundle_path) as local_bundle_path:
                return func(local_bundle_path, *args)

        return func(bundle_path, *args)

    return wrapper


@resolve_remote_bundle
def load_saved_bundle_config(bundle_path) -> "SavedBundleConfig":
    try:
        return SavedBundleConfig.load(os.path.join(bundle_path, "bentoml.yml"))
    except FileNotFoundError:
        raise BentoMLException(
            "BentoML can't locate config file 'bentoml.yml'"
            " in saved bundle in path: {}".format(bundle_path)
        )


def load_bento_service_metadata(bundle_path: str) -> "BentoServiceMetadata":
    return load_saved_bundle_config(bundle_path).get_bento_service_metadata_pb()


def _find_module_file(bundle_path, service_name, module_file):
    # Simply join full path when module_file is just a file name,
    # e.g. module_file=="iris_classifier.py"
    module_file_path = os.path.join(bundle_path, service_name, module_file)
    if not os.path.isfile(module_file_path):
        # Try loading without service_name prefix, for loading from a installed PyPi
        module_file_path = os.path.join(bundle_path, module_file)

    # When module_file is located in sub directory
    # e.g. module_file=="foo/bar/iris_classifier.py"
    # This needs to handle the path differences between posix and windows platform:
    if not os.path.isfile(module_file_path):
        if sys.platform == "win32":
            # Try load a saved bundle created from posix platform on windows
            module_file_path = os.path.join(
                bundle_path, service_name, str(PurePosixPath(module_file))
            )
            if not os.path.isfile(module_file_path):
                module_file_path = os.path.join(
                    bundle_path, str(PurePosixPath(module_file))
                )
        else:
            # Try load a saved bundle created from windows platform on posix
            module_file_path = os.path.join(
                bundle_path, service_name, PureWindowsPath(module_file).as_posix()
            )
            if not os.path.isfile(module_file_path):
                module_file_path = os.path.join(
                    bundle_path, PureWindowsPath(module_file).as_posix()
                )

    if not os.path.isfile(module_file_path):
        raise BentoMLException(
            "Can not locate module_file {} in saved bundle {}".format(
                module_file, bundle_path
            )
        )

    return module_file_path


@resolve_remote_bundle
def load_bento_service_class(bundle_path):
    """
    Load a BentoService class from saved bundle in given path

    :param bundle_path: A path to Bento files generated from BentoService#save,
        #save_to_dir, or the path to pip installed BentoService directory
    :return: BentoService class
    """
    config = load_saved_bundle_config(bundle_path)
    metadata = config["metadata"]

    # Find and load target module containing BentoService class from given path
    module_file_path = _find_module_file(
        bundle_path, metadata["service_name"], metadata["module_file"]
    )

    # Prepend bundle_path to sys.path for loading extra python dependencies
    sys.path.insert(0, bundle_path)
    sys.path.insert(0, os.path.join(bundle_path, metadata["service_name"]))
    # Include zipimport modules
    zipimport_dir = os.path.join(bundle_path, metadata["service_name"], ZIPIMPORT_DIR)
    if os.path.exists(zipimport_dir):
        for p in os.listdir(zipimport_dir):
            logger.debug('adding %s to sys.path', p)
            sys.path.insert(0, os.path.join(zipimport_dir, p))

    module_name = metadata["module_name"]
    if module_name in sys.modules:
        logger.warning(
            "Module `%s` already loaded, using existing imported module.", module_name
        )
        module = sys.modules[module_name]
    elif sys.version_info >= (3, 5):
        import importlib.util

        spec = importlib.util.spec_from_file_location(module_name, module_file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    elif sys.version_info >= (3, 3):
        from importlib.machinery import SourceFileLoader

        # pylint:disable=deprecated-method
        module = SourceFileLoader(module_name, module_file_path).load_module(
            module_name
        )
        # pylint:enable=deprecated-method
    else:
        raise BentoMLException("BentoML requires Python 3.4 and above")

    # Remove bundle_path from sys.path to avoid import naming conflicts
    sys.path.remove(bundle_path)

    model_service_class = module.__getattribute__(metadata["service_name"])
    # Set _bento_service_bundle_path, where BentoService will load its artifacts
    model_service_class._bento_service_bundle_path = bundle_path
    # Set cls._version, service instance can access it via svc.version
    model_service_class._bento_service_bundle_version = metadata["service_version"]

    if (
        model_service_class._env
        and model_service_class._env._requirements_txt_file is not None
    ):
        # Load `requirement.txt` from bundle directory instead of the user-provided
        # file path, which may only available during the bundle save process
        model_service_class._env._requirements_txt_file = os.path.join(
            bundle_path, "requirements.txt"
        )

    return model_service_class


@resolve_remote_bundle
def safe_retrieve(bundle_path: str, target_dir: str):
    """Safely retrieve bento service to local path

    Args:
        bundle_path (:obj:`str`):
            The path that contains saved BentoService bundle, supporting
            both local file path and s3 path
        target_dir (:obj:`str`):
            Where the service contents should end up.

    Returns:
        :obj:`str`: location of safe local path
    """
    shutil.copytree(bundle_path, target_dir)


@resolve_remote_bundle
def load_from_dir(bundle_path):
    """Load bento service from local file path or s3 path

    Args:
        bundle_path (str): The path that contains saved BentoService bundle,
            supporting both local file path and s3 path

    Returns:
        bentoml.service.BentoService: a loaded BentoService instance
    """
    svc_cls = load_bento_service_class(bundle_path)
    return svc_cls()


@resolve_remote_bundle
def load_bento_service_api(bundle_path, api_name=None):
    bento_service = load_from_dir(bundle_path)
    return bento_service.get_inference_api(api_name)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import functools
import inspect
from io import StringIO
import os
import socket
import tarfile
from typing import Optional
from urllib.parse import urlparse, uses_netloc, uses_params, uses_relative

from google.protobuf.message import Message

from bentoml.utils.gcs import is_gcs_url
from bentoml.utils.lazy_loader import LazyLoader
from bentoml.utils.s3 import is_s3_url

_VALID_URLS = set(uses_relative + uses_netloc + uses_params)
_VALID_URLS.discard("")


__all__ = [
    "cached_property",
    "reserve_free_port",
    "is_url",
    "dump_to_yaml_str",
    "pb_to_yaml",
    "ProtoMessageToDict",
    "status_pb_to_error_code_and_message",
    "catch_exceptions",
    "cached_contextmanager",
    "resolve_bento_bundle_uri",
    "is_gcs_url",
    "is_s3_url",
    "archive_directory_to_tar",
    "resolve_bundle_path",
]

yatai_proto = LazyLoader("yatai_proto", globals(), "bentoml.yatai.proto")

DEFAULT_CHUNK_SIZE = 1024 * 8  # 8kb


class _Missing(object):
    def __repr__(self):
        return "no value"

    def __reduce__(self):
        return "_missing"


_missing = _Missing()


class cached_property(property):
    """A decorator that converts a function into a lazy property. The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value:

    .. code-block:: python

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """

    # implementation detail: A subclass of python's builtin property
    # decorator, we override __get__ to check for a cached value. If one
    # chooses to invoke __get__ by hand the property will still work as
    # expected because the lookup logic is replicated in __get__ for
    # manual invocation.

    def __init__(
        self, func, name=None, doc=None
    ):  # pylint:disable=super-init-not-called
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):  # pylint:disable=redefined-builtin
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class cached_contextmanager:
    """
    Just like contextlib.contextmanager, but will cache the yield value for the same
    arguments. When one instance of the contextmanager exits, the cache value will
    also be poped.

    Example Usage:
    (To reuse the container based on the same image)

    >>> @cached_contextmanager("{docker_image.id}")
    >>> def start_docker_container_from_image(docker_image, timeout=60):
    >>>     container = ...
    >>>     yield container
    >>>     container.stop()
    """

    def __init__(self, cache_key_template=None):
        self._cache_key_template = cache_key_template
        self._cache = {}

    def __call__(self, func):
        func_m = contextlib.contextmanager(func)

        @contextlib.contextmanager
        @functools.wraps(func)
        def _func(*args, **kwargs):
            bound_args = inspect.signature(func).bind(*args, **kwargs)
            bound_args.apply_defaults()
            if self._cache_key_template:
                cache_key = self._cache_key_template.format(**bound_args.arguments)
            else:
                cache_key = tuple(bound_args.arguments.values())
            if cache_key in self._cache:
                yield self._cache[cache_key]
            else:
                with func_m(*args, **kwargs) as value:
                    self._cache[cache_key] = value
                    yield value
                    self._cache.pop(cache_key)

        return _func


@contextlib.contextmanager
def reserve_free_port(host="localhost"):
    """
    detect free port and reserve until exit the context
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, 0))
    port = sock.getsockname()[1]
    yield port
    sock.close()


def is_url(url: str) -> bool:
    try:
        return urlparse(url).scheme in _VALID_URLS
    except Exception:  # pylint:disable=broad-except
        return False


def dump_to_yaml_str(yaml_dict):
    from bentoml.utils.ruamel_yaml import YAML

    yaml = YAML()
    string_io = StringIO()
    yaml.dump(yaml_dict, string_io)
    return string_io.getvalue()


def pb_to_yaml(message: Message) -> str:
    from google.protobuf.json_format import MessageToDict

    message_dict = MessageToDict(message)
    return dump_to_yaml_str(message_dict)


def ProtoMessageToDict(protobuf_msg: Message, **kwargs) -> object:
    from google.protobuf.json_format import MessageToDict

    if "preserving_proto_field_name" not in kwargs:
        kwargs["preserving_proto_field_name"] = True

    return MessageToDict(protobuf_msg, **kwargs)


# This function assume the status is not status.OK
def status_pb_to_error_code_and_message(pb_status) -> (int, str):
    from bentoml.yatai.proto import status_pb2

    assert pb_status.status_code != status_pb2.Status.OK
    error_code = status_pb2.Status.Code.Name(pb_status.status_code)
    error_message = pb_status.error_message
    return error_code, error_message


class catch_exceptions(object):
    def __init__(self, exceptions, fallback=None):
        self.exceptions = exceptions
        self.fallback = fallback

    def __call__(self, func):
        @functools.wraps(func)
        def _(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except self.exceptions:
                return self.fallback

        return _


def resolve_bundle_path(
    bento: str,
    pip_installed_bundle_path: Optional[str] = None,
    yatai_url: Optional[str] = None,
) -> str:
    from bentoml.exceptions import BentoMLException
    from bentoml.yatai.client import get_yatai_client

    if pip_installed_bundle_path:
        assert (
            bento is None
        ), "pip installed BentoService commands should not have Bento argument"
        return pip_installed_bundle_path

    if os.path.isdir(bento) or is_s3_url(bento) or is_gcs_url(bento):
        # saved_bundle already support loading local, s3 path and gcs path
        return bento

    elif ":" in bento:
        # assuming passing in BentoService in the form of Name:Version tag
        yatai_client = get_yatai_client(yatai_url)
        bento_pb = yatai_client.repository.get(bento)
        return resolve_bento_bundle_uri(bento_pb)
    else:
        raise BentoMLException(
            f'BentoService "{bento}" not found - either specify the file path of '
            f"the BentoService saved bundle, or the BentoService id in the form of "
            f'"name:version"'
        )


def get_default_yatai_client():
    from bentoml.yatai.client import YataiClient

    return YataiClient()


def resolve_bento_bundle_uri(bento_pb):
    if bento_pb.uri.s3_presigned_url:
        # Use s3 presigned URL for downloading the repository if it is presented
        return bento_pb.uri.s3_presigned_url
    if bento_pb.uri.gcs_presigned_url:
        return bento_pb.uri.gcs_presigned_url
    else:
        return bento_pb.uri.uri


def archive_directory_to_tar(
    source_dir: str, tarfile_dir: str, tarfile_name: str
) -> (str, str):
    file_name = f'{tarfile_name}.tar'
    tarfile_path = os.path.join(tarfile_dir, file_name)
    with tarfile.open(tarfile_path, mode="w:gz") as tar:
        # Use arcname to prevent storing the full path to the bundle,
        # from source_dir/path/to/file to path/to/file
        tar.add(source_dir, arcname='')
    return tarfile_path, file_name



import os
import re

from bentoml.service import BentoServiceArtifact


JSON_ARTIFACT_EXTENSION = ".json"


class JSONArtifact(BentoServiceArtifact):
    """Abstraction for saving/loading objects to/from JSON files.

    Args:
        name (str): Name of the artifact
        encoding (:obj:`str`, optional): The encoding will be used for saving/loading
            text. Defaults to "utf8"
        json_module (module|object, optional): Namespace/object implementing `loads()`
            and `dumps()` methods for serializing/deserializing to/from JSON string.
            Defaults to stdlib's json module.
    """

    def __init__(self, name, encoding="utf-8", json_module=None):
        super().__init__(name)
        self._content = None
        self._json_dumps_kwargs = None

        self._encoding = encoding
        if json_module:
            self.json_module = json_module
        else:
            import json

            self.json_module = json

    def _file_path(self, base_path):
        return os.path.join(
            base_path,
            re.sub("[^-a-zA-Z0-9_.() ]+", "", self.name) + JSON_ARTIFACT_EXTENSION,
        )

    def load(self, path):
        with open(self._file_path(path), "rt", encoding=self._encoding) as fp:
            content = self.json_module.loads(fp.read())
        return self.pack(content)

    def pack(
        self, content, metadata=None, **json_dumps_kwargs
    ):  # pylint:disable=arguments-renamed
        self._content = content
        self._json_dumps_kwargs = json_dumps_kwargs
        return self

    def get(self):
        return self._content

    def save(self, dst):
        with open(self._file_path(dst), "wt", encoding=self._encoding) as fp:
            fp.write(self.json_module.dumps(self._content, **self._json_dumps_kwargs))



MAX_RESOURCE_GROUP_NAME_LENGTH = 90
MAX_STORAGE_ACCOUNT_NAME_LENGTH = 24
MAX_FUNCTION_NAME_LENGTH = 60
MAX_CONTAINER_REGISTRY_NAME_LENGTH = 50

# https://docs.microsoft.com/en-us/azure/azure-functions/functions-premium-plan
AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS = ['EP1', 'EP2', 'EP3']
# https://docs.microsoft.com/en-us/java/api/com.microsoft.azure.functions.annotation.httptrigger.authlevel?view=azure-java-stable
AZURE_FUNCTIONS_AUTH_LEVELS = ['anonymous', 'function', 'admin']

DEFAULT_MIN_INSTANCE_COUNT = 1
DEFAULT_MAX_BURST = 20
DEFAULT_PREMIUM_PLAN_SKU = AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS[0]
DEFAULT_FUNCTION_AUTH_LEVEL = AZURE_FUNCTIONS_AUTH_LEVELS[0]



import logging
import os

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv

logger = logging.getLogger(__name__)


COREML_MODEL_FILE_EXTENTION = ".mlmodel"


class CoreMLModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading coreml.models.MLModel objects
    with coremltools.models.MLModel.save and coremltools.models.MLModel(path).

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: coremltools package required for CoreMLModelArtifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            coremltools.models.MLModel

    Example usage:

    >>> import coremltools as ct
    >>> import torch.nn as nn
    >>>
    >>> class Net(nn.Module):
    >>>     def __init__(self):
    >>>         super(Net, self).__init__()
    >>>         ...
    >>>
    >>>     def forward(self, x):
    >>>         ...
    >>>
    >>> net = Net()
    >>> # Train model with data, then convert to CoreML.
    >>> model = ct.convert(net, ...)
    >>>
    >>> import bentoml
    >>> import PIL.Image
    >>> from bentoml.adapters import ImageInput
    >>> from bentoml.frameworks.coreml import CoreMLModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([CoreMLModelArtifact('model')])
    >>> class CoreMLModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=ImageInput(), batch=True)
    >>>     def predict(self, imgs):
    >>>         outputs = [self.artifacts.model(PIL.Image.fromarray(_.astype("uint8")))
    >>>                    for img in imgs]
    >>>         return outputs
    >>>
    >>>
    >>> svc = CoreMLModelService()
    >>>
    >>> # Pytorch model can be packed directly.
    >>> svc.pack('model', model)
    """

    def __init__(self, name: str):
        super().__init__(name)
        self._model = None

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name + COREML_MODEL_FILE_EXTENTION)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        try:
            import coremltools
        except ImportError:
            raise MissingDependencyException(
                "coremltools>=4.0b2 package is required to use CoreMLModelArtifact"
            )

        if not isinstance(model, coremltools.models.MLModel):
            raise InvalidArgument(
                "CoreMLModelArtifact can only pack type 'coremltools.models.MLModel'"
            )

        self._model = model
        return self

    def load(self, path):
        try:
            import coremltools
        except ImportError:
            raise MissingDependencyException(
                "coremltools package is required to use CoreMLModelArtifact"
            )

        model = coremltools.models.MLModel(self._file_path(path))

        if not isinstance(model, coremltools.models.MLModel):
            raise InvalidArgument(
                "Expecting CoreMLModelArtifact loaded object type to be "
                "'coremltools.models.MLModel' but actually it is {}".format(type(model))
            )

        return self.pack(model)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['coremltools>=4.0b2'])

    def get(self):
        return self._model

    def save(self, dst):
        self._model.save(self._file_path(dst))



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from pathlib import Path
import stat
from sys import version_info
from typing import List

from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException
from bentoml.saved_bundle.pip_pkg import (
    EPP_NO_ERROR,
    EPP_PKG_NOT_EXIST,
    EPP_PKG_VERSION_MISMATCH,
    get_pkg_version,
    verify_pkg,
)
from bentoml.utils.ruamel_yaml import YAML
from bentoml.utils import cached_property


logger = logging.getLogger(__name__)

PYTHON_SUPPORTED_VERSIONS = ["3.6", "3.7", "3.8"]

PYTHON_MINOR_VERSION = "{major}.{minor}".format(
    major=version_info.major, minor=version_info.minor
)

PYTHON_VERSION = "{minor_version}.{micro}".format(
    minor_version=PYTHON_MINOR_VERSION, micro=version_info.micro
)


DEFAULT_CONDA_ENV_BASE_YAML = """
name: bentoml-default-conda-env
dependencies: []
"""


class CondaEnv(object):
    """A wrapper around conda environment settings file, allows adding/removing
    conda or pip dependencies to env configuration, and supports load/export those
    settings from/to yaml files. The generated file is the same format as yaml file
    generated from `conda env export` command.
    """

    def __init__(
        self,
        name: str = None,
        channels: List[str] = None,
        dependencies: List[str] = None,
        default_env_yaml_file: str = None,
        override_channels: bool = False,
    ):
        self._yaml = YAML()
        self._yaml.default_flow_style = False
        self.default_env_yaml_file = default_env_yaml_file

        if name:
            self.set_name(name)

        if override_channels and channels is None:
            raise BentoMLException(
                "No `conda_channels` provided while override_channels=True"
            )

        if channels:
            self.set_channels(channels, override_channels)

        if dependencies:
            self.add_conda_dependencies(dependencies)

    @cached_property
    def _conda_env(self):
        if self.default_env_yaml_file:
            env_yml_file = Path(self.default_env_yaml_file)
            if not env_yml_file.is_file():
                raise BentoMLException(
                    f"Can not find conda environment config yaml file at: "
                    f"`{self.default_env_yaml_file}`"
                )
            return self._yaml.load(env_yml_file)
        else:
            return self._yaml.load(DEFAULT_CONDA_ENV_BASE_YAML)

    def set_name(self, name):
        self._conda_env["name"] = name

    def add_conda_dependencies(self, conda_dependencies: List[str]):
        # BentoML uses conda's channel_priority=strict option by default
        # Adding `dependencies` to beginning of the list to take priority over the
        # existing conda channels
        self._conda_env["dependencies"] = (
            conda_dependencies + self._conda_env["dependencies"]
        )

    def set_channels(self, channels: List[str], override_channels=False):
        if override_channels and "nodefaults" not in channels:
            channels.append("nodefaults")
        self._conda_env["channels"] = channels

    def write_to_yaml_file(self, filepath):
        with open(filepath, 'wb') as output_yaml:
            self._yaml.dump(self._conda_env, output_yaml)


class BentoServiceEnv(object):
    """Defines all aspect of the system environment requirements for a custom
    BentoService to be used. This includes:


    Args:
        pip_packages: list of pip_packages required, specified by package name
            or with specified version `{package_name}=={package_version}`
        pip_index_url: passing down to pip install --index-url option
        pip_trusted_host: passing down to pip install --trusted-host option
        pip_extra_index_url: passing down to pip install --extra-index-url option
        infer_pip_packages: Turn on to automatically find all the required
            pip dependencies and pin their version
        requirements_txt_file: path to the requirements.txt where pip dependencies
            are explicitly specified, with ideally pinned versions
        conda_channels: list of extra conda channels other than default channels to be
            used. This is equivalent to passing the --channels to conda commands.
            If the `conda_env_yml_file` is specified, this will override the `channels`
            section of the env yml file
        conda_override_channels: ensures that conda searches only your specified
            channel and no other channels, such as default channels.
            This is equivalent to passing the --override-channels option to conda
            commands, or adding `nodefaults` to the `channels` in the environment.yml
        conda_overwrite_channels: aliases to `override_channels`
        conda_dependencies: list of conda dependencies required
        conda_env_yml_file: use a pre-defined conda environment yml file
        setup_sh: user defined setup bash script, it is executed in docker build time
        docker_base_image: used when generating Dockerfile in saved bundle
        zipimport_archives: used to list zipimport archives
    """

    @inject
    def __init__(
        self,
        pip_packages: List[str] = None,
        pip_index_url: str = None,
        pip_trusted_host: str = None,
        pip_extra_index_url: str = None,
        infer_pip_packages: bool = False,
        requirements_txt_file: str = None,
        conda_channels: List[str] = None,
        conda_overwrite_channels: bool = False,
        conda_override_channels: bool = False,
        conda_dependencies: List[str] = None,
        conda_env_yml_file: str = None,
        setup_sh: str = None,
        docker_base_image: str = None,
        default_docker_base_image: str = Provide[
            BentoMLContainer.config.bento_bundle.default_docker_base_image
        ],
        zipimport_archives: List[str] = None,
        bentoml_deployment_version: str = Provide[
            BentoMLContainer.bento_bundle_deployment_version
        ],
    ):
        self._python_version = PYTHON_VERSION
        self._pip_index_url = pip_index_url
        self._pip_trusted_host = pip_trusted_host
        self._pip_extra_index_url = pip_extra_index_url
        self._requirements_txt_file = requirements_txt_file
        self._requirements_txt_content = None

        self._conda_env = CondaEnv(
            channels=conda_channels,
            dependencies=conda_dependencies,
            default_env_yaml_file=conda_env_yml_file,
            override_channels=conda_overwrite_channels or conda_override_channels,
        )

        self._pip_packages = {}

        # add BentoML to pip packages list
        bentoml_deploy_version = bentoml_deployment_version
        self.add_pip_package("bentoml=={}".format(bentoml_deploy_version))

        if pip_packages:
            self.add_pip_packages(pip_packages)

        self._infer_pip_packages = infer_pip_packages

        self.set_setup_sh(setup_sh)

        if docker_base_image:
            logger.info(
                f"Using user specified docker base image: `{docker_base_image}`, user"
                f"must make sure that the base image either has Python "
                f"{PYTHON_MINOR_VERSION} or conda installed."
            )
            self._docker_base_image = docker_base_image
        elif default_docker_base_image:
            logger.info(
                f"Using default docker base image: `{docker_base_image}` specified in"
                f"BentoML config file or env var. User must make sure that the docker "
                f"base image either has Python {PYTHON_MINOR_VERSION} or conda "
                f"installed."
            )
            self._docker_base_image = default_docker_base_image
        else:
            if PYTHON_MINOR_VERSION not in PYTHON_SUPPORTED_VERSIONS:
                self._docker_base_image = (
                    f"bentoml/model-server:{bentoml_deploy_version}"
                )

                logger.warning(
                    f"Python {PYTHON_VERSION} found in current environment is not "
                    f"officially supported by BentoML. The docker base image used is"
                    f"'{self._docker_base_image}' which will use conda to install "
                    f"Python {PYTHON_VERSION} in the build process. Supported Python "
                    f"versions are: f{', '.join(PYTHON_SUPPORTED_VERSIONS)}"
                )
            else:
                # e.g. bentoml/model-server:0.8.6-py37
                self._docker_base_image = (
                    f"bentoml/model-server:"
                    f"{bentoml_deploy_version}-"
                    f"py{PYTHON_MINOR_VERSION.replace('.', '')}"
                )

                logger.debug(
                    f"Using BentoML default docker base image "
                    f"'{self._docker_base_image}'"
                )
        self._zipimport_archives = zipimport_archives

    def add_conda_dependencies(self, conda_dependencies: List[str]):
        self._conda_env.add_conda_dependencies(conda_dependencies)

    def add_pip_packages(self, pip_packages: List[str]):
        for pip_package in pip_packages:
            self.add_pip_package(pip_package)

    def add_pip_package(self, pip_package: str):
        # str must be a valid pip requirement specifier
        # https://pip.pypa.io/en/stable/reference/pip_install/#requirement-specifiers
        from pkg_resources import Requirement

        package_req = Requirement(pip_package)
        self._add_pip_package_requirement(package_req)

    def _add_pip_package_requirement(self, pkg_req):
        from pkg_resources import Requirement

        if pkg_req.name in self._pip_packages:
            if (
                pkg_req.specs
                and pkg_req.specs != self._pip_packages[pkg_req.name].specs
            ):
                logger.warning(
                    f"Overwriting existing pip package requirement "
                    f"'{self._pip_packages[pkg_req.name]}' to '{pkg_req}'"
                )
            else:
                logger.debug(f"pip package requirement '{pkg_req}' already exist")
                return

        verification_result = verify_pkg(pkg_req)
        if verification_result == EPP_PKG_NOT_EXIST:
            logger.warning(
                f"pip package requirement `{str(pkg_req)}` not found in current "
                f"python environment"
            )
        elif verification_result == EPP_PKG_VERSION_MISMATCH:
            logger.warning(
                f"pip package requirement `{str(pkg_req)}` does not match the "
                f"version installed in current python environment"
            )

        if verification_result == EPP_NO_ERROR and not pkg_req.specs:
            # pin the current version when there's no version spec found
            pkg_version = get_pkg_version(pkg_req.name)
            if pkg_version:
                pkg_req = Requirement(f"{pkg_req.name}=={pkg_version}")

        self._pip_packages[pkg_req.name] = pkg_req

    def set_setup_sh(self, setup_sh_path_or_content):
        self._setup_sh = None

        if setup_sh_path_or_content:
            setup_sh_file = Path(setup_sh_path_or_content)
        else:
            return

        if setup_sh_file.is_file():
            with setup_sh_file.open("rb") as f:
                self._setup_sh = f.read()
        else:
            self._setup_sh = setup_sh_path_or_content.encode("utf-8")

    def requirements_txt_content(self):
        if self._requirements_txt_content:
            # This branch handles the case when the _requirements_txt_content is
            # previously set by the BentoBundle loader
            return self._requirements_txt_content

        if not self._requirements_txt_file:
            raise BentoMLException("requirement txt file not specified")

        if not self._requirements_txt_content:
            req_txt_file = Path(self._requirements_txt_file)
            if not req_txt_file.is_file():
                raise BentoMLException(
                    f"requirement txt file not found at '{self._requirements_txt_file}'"
                )
            self._requirements_txt_content = req_txt_file.read_text()

        return self._requirements_txt_content

    def infer_pip_packages(self, bento_service):
        if self._infer_pip_packages:
            dependencies_map = bento_service.infer_pip_dependencies_map()
            for pkg_name, pkg_version in dependencies_map.items():
                if (
                    pkg_name in self._pip_packages
                    and self._pip_packages[pkg_name].specs
                ):
                    # package exist and already have version specs associated
                    continue
                else:
                    # pin current version if the package has not been added
                    self.add_pip_package(f"{pkg_name}=={pkg_version}")

    def save(self, path):
        conda_yml_file = os.path.join(path, "environment.yml")
        self._conda_env.write_to_yaml_file(conda_yml_file)

        with open(os.path.join(path, "python_version"), "wb") as f:
            f.write(self._python_version.encode("utf-8"))

        # If requirements.txt is specified, other requirements will
        # NOT be parsed (#1421)
        requirements_txt_file = os.path.join(path, "requirements.txt")
        with open(requirements_txt_file, "wb") as f:
            if self._requirements_txt_file:
                f.write(self.requirements_txt_content().encode("utf-8"))
            else:
                if self._pip_index_url:
                    f.write(f"--index-url={self._pip_index_url}\n".encode("utf-8"))
                if self._pip_trusted_host:
                    f.write(
                        f"--trusted-host={self._pip_trusted_host}\n".encode("utf-8")
                    )
                if self._pip_extra_index_url:
                    f.write(
                        f"--extra-index-url={self._pip_extra_index_url}\n".encode(
                            "utf-8"
                        )
                    )
                pip_content = '\n'.join(
                    [str(pkg_req) for pkg_req in self._pip_packages.values()]
                ).encode("utf-8")
                f.write(pip_content)

        if self._setup_sh:
            setup_sh_file = os.path.join(path, "setup.sh")
            with open(setup_sh_file, "wb") as f:
                f.write(self._setup_sh)

            # chmod +x setup.sh
            st = os.stat(setup_sh_file)
            os.chmod(setup_sh_file, st.st_mode | stat.S_IEXEC)

    def to_dict(self):
        env_dict = dict()

        if self._setup_sh:
            env_dict["setup_sh"] = self._setup_sh

        if self._requirements_txt_file:
            env_dict["requirements_txt"] = self.requirements_txt_content()
        elif self._pip_packages:
            # pip_packages are ignored when the requirements_txt_file parameter is
            # specified by the user
            env_dict["pip_packages"] = [
                str(pkg_req) for pkg_req in self._pip_packages.values()
            ]

        env_dict["conda_env"] = self._conda_env._conda_env
        env_dict["python_version"] = self._python_version
        env_dict["docker_base_image"] = self._docker_base_image
        return env_dict



import logging
import os

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv

logger = logging.getLogger(__name__)


class SpacyModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading spacy models with Language.to_disk and
    spacy.util.load_model methods

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: spacy package is required for SpacyModelArtifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            spacy.language.Language

    Example usage:

    >>> import spacy
    >>> # TRAIN MODEL WITH DATA
    >>> TRAIN_DATA = [
    >>>     ("Uber blew through $1 million a week", {"entities": [(0, 4, "ORG")]}),
    >>>     ("Google rebrands its business apps", {"entities": [(0, 6, "ORG")]})]
    >>>
    >>> nlp = spacy.blank("en")
    >>> optimizer = nlp.begin_training()
    >>> for i in range(20):
    >>>     random.shuffle(TRAIN_DATA)
    >>>     for text, annotations in TRAIN_DATA:
    >>>         nlp.update([text], [annotations], sgd=optimizer)
    >>> nlp.to_disk("/model")
    >>>
    >>> import bentoml
    >>> from bentoml.adapters import JsonInput
    >>> from bentoml.frameworks.spacy import SpacyModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([SpacyModelArtifact('nlp')])
    >>> class SpacyModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=JsonInput(), batch=False)
    >>>     def predict(self, parsed_json):
    >>>         output = self.artifacts.nlp(parsed_json['text'])
    >>>         return output
    >>>
    >>>
    >>> svc = SpacyModelService()
    >>>
    >>> # Spacy model can be packed directly.
    >>> svc.pack('nlp', nlp)
    """

    def __init__(self, name):
        super().__init__(name)

        self._model = None

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        try:
            import spacy
        except ImportError:
            raise MissingDependencyException(
                "spacy package is required to use SpacyModelArtifact"
            )

        if not isinstance(model, spacy.language.Language):
            raise InvalidArgument(
                "SpacyModelArtifact can only pack type 'spacy.language.Language'"
            )

        self._model = model
        return self

    def load(self, path):
        try:
            import spacy
        except ImportError:
            raise MissingDependencyException(
                "spacy package is required to use SpacyModelArtifact"
            )

        model = spacy.util.load_model(self._file_path(path))

        if not isinstance(model, spacy.language.Language):
            raise InvalidArgument(
                "Expecting SpacyModelArtifact loaded object type to be "
                "'spacy.language.Language' but actually it is {}".format(type(model))
            )

        return self.pack(model)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['spacy'])

    def get(self):
        return self._model

    def save(self, dst):
        path = self._file_path(dst)
        return self._model.to_disk(path)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

import versioneer

with open("README.md", "r", encoding="utf8") as f:
    long_description = f.read()

install_requires = [
    "aiohttp",
    "aiohttp_cors==0.7.0",
    "alembic",
    "urllib3<=1.25.11",
    "boto3",
    "cerberus",
    "certifi",
    "click>=7.0",
    "configparser",
    "deepmerge",
    "docker",
    "flask",
    "grpcio",
    "gunicorn",
    "humanfriendly",
    "numpy",
    "packaging",
    "prometheus_client",
    "protobuf>=3.8.0",
    "psutil",
    # python-dateutil required by pandas and boto3, this makes sure the version
    # works for both
    "python-dateutil>=2.7.3,<3.0.0",
    "python-json-logger",
    "requests",
    "ruamel.yaml>=0.15.0",
    "schema",
    "sqlalchemy-utils<0.36.8",
    "sqlalchemy>=1.3.0,<1.4.0",
    "tabulate",
    'contextvars;python_version < "3.7"',
    'dataclasses;python_version < "3.7"',
    "chardet",
    "simple-di",
]

yatai_service_requires = [
    "grpcio~=1.34.0",  # match the grpcio-tools version used in yatai docker image
    "google-cloud-storage",
    "azure-cli",
    "aws-sam-cli==0.33.1",
    "psycopg2",
    "psycopg2-binary",
]

model_server_requires = [
    "opentracing",
    "py_zipkin",
    "jaeger_client",
]

test_requires = [
    "idna<=2.8",  # for moto
    "ecdsa==0.14",  # for moto
    "black==19.10b0",
    "codecov",
    "coverage>=4.4",
    "flake8>=3.8.2",
    "imageio>=2.5.0",
    "mock>=2.0.0",
    "moto==1.3.14",
    "pandas",
    "pylint>=2.9.3",
    "pytest-cov>=2.7.1",
    "pytest>=5.4.0",
    "pytest-asyncio",
    "parameterized",
    "scikit-learn",
]

dev_requires = [
    "flake8>=3.8.2",
    "gitpython>=2.0.2",
    # grpcio-tools version must be kept in sync with the version used in
    # `protos/generate-docker.sh` script
    "grpcio-tools~=1.34.0",
    "grpcio-reflection~=1.34.0",
    "pylint>=2.5.2",
    "setuptools",
    "tox-conda>=0.2.0",
    "tox>=3.12.1",
    "twine",
] + test_requires

docs_requires = [
    "recommonmark",
    "sphinx<=3.5.4",
    "sphinx-click",
    "sphinx_rtd_theme",
    "sphinxcontrib-fulltoc",
    "sphinxcontrib-spelling",
    "sphinx_copybutton",
    "pyenchant",
]

types_requires = [
    "mypy-protobuf",
    "types-click",
    "types-protobuf>=0.1.14",
    "grpc-stubs",
]

dev_all = install_requires + dev_requires + docs_requires + types_requires

extras_require = {
    "dev": dev_all,
    "test": test_requires,
    "yatai_service": yatai_service_requires,
    "model_server": model_server_requires,
    "doc_builder": docs_requires,  # 'doc_builder' is required by readthedocs.io
    "types_stub": types_requires,
}

setuptools.setup(
    name="BentoML",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="bentoml.org",
    author_email="contact@bentoml.ai",
    description="A framework for machine learning model serving",
    long_description=long_description,
    license="Apache License 2.0",
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    extras_require=extras_require,
    url="https://github.com/bentoml/BentoML",
    packages=setuptools.find_packages(exclude=["tests*"]),
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.6.1",
    entry_points={"console_scripts": ["bentoml=bentoml:commandline_interface"]},
    project_urls={
        "Bug Reports": "https://github.com/bentoml/BentoML/issues",
        "BentoML User Slack Group": "https://bit.ly/2N5IpbB",
        "Source Code": "https://github.com/bentoml/BentoML",
    },
    include_package_data=True,  # Required for '.cfg' files under bentoml/config
)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Iterable, Iterator, Sequence, Tuple

import chardet

from bentoml.adapters.base_input import BaseInputAdapter
from bentoml.adapters.utils import decompress_gzip_request
from bentoml.types import AwsLambdaEvent, FileLike, HTTPRequest, InferenceTask

ApiFuncArgs = Tuple[
    Sequence[str],
]


class StringInput(BaseInputAdapter):
    """
    Convert various inputs(HTTP, Aws Lambda or CLI) to strings(list of str), passing it
    to API functions.

    Errors
        400 UnicodeDecodeError

        400 Unsupported charset

    Example Request

    >>> curl -i \\
    >>>    --header "Content-Type: text/plain; charset=utf-8" \\
    >>>    --request POST \\
    >>>    --data 'best movie ever' \\
    >>>    localhost:5000/predict
    """

    BATCH_MODE_SUPPORTED = True

    @decompress_gzip_request
    def from_http_request(self, req: HTTPRequest) -> InferenceTask[str]:
        if req.headers.content_type == 'multipart/form-data':
            _, _, files = HTTPRequest.parse_form_data(req)
            if len(files) != 1:
                return InferenceTask().discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} accepts one text file "
                    "at a time",
                )
            input_file = next(iter(files.values()))
            bytes_ = input_file.read()
            charset = chardet.detect(bytes_)['encoding'] or "utf-8"
        else:
            bytes_ = req.body
            charset = req.headers.charset or "utf-8"
        try:
            return InferenceTask(http_headers=req.headers, data=bytes_.decode(charset),)
        except UnicodeDecodeError:
            return InferenceTask().discard(
                http_status=400,
                err_msg=f"{self.__class__.__name__}: UnicodeDecodeError for {req.body}",
            )
        except LookupError:
            return InferenceTask().discard(
                http_status=400,
                err_msg=f"{self.__class__.__name__}: Unsupported charset {req.charset}",
            )

    def from_aws_lambda_event(self, event: AwsLambdaEvent) -> InferenceTask[str]:
        return InferenceTask(aws_lambda_event=event, data=event.get('body', ""),)

    def from_cli(self, cli_args: Tuple[str]) -> Iterator[InferenceTask[str]]:
        import argparse

        parser = argparse.ArgumentParser()
        input_g = parser.add_mutually_exclusive_group(required=True)
        input_g.add_argument('--input', nargs="+", type=str)
        input_g.add_argument('--input-file', nargs="+")

        parsed_args, _ = parser.parse_known_args(list(cli_args))

        for t in self.from_inference_job(
            input_=parsed_args.input, input_file=parsed_args.input_file,
        ):
            t.cli_args = cli_args
            yield t

    def from_inference_job(  # pylint: disable=arguments-differ
        self, input_=None, input_file=None, **extra_args,
    ) -> Iterator[InferenceTask[str]]:
        '''
        Generate InferenceTask from calling bentom_svc.run(input_=None, input_file=None)

        Parameters
        ----------
        input_ : str
            The input value

        input_file : str
            The URI/path of the input file

        extra_args : dict
            Additional parameters
        '''
        if input_file is not None:
            for d in input_file:
                uri = pathlib.Path(d).absolute().as_uri()
                bytes_ = FileLike(uri=uri).read()
                try:
                    charset = chardet.detect(bytes_)['encoding'] or "utf-8"
                    yield InferenceTask(
                        inference_job_args=extra_args, data=bytes_.decode(charset),
                    )
                except UnicodeDecodeError:
                    yield InferenceTask().discard(
                        http_status=400,
                        err_msg=f"{self.__class__.__name__}: "
                        f"Try decoding with {charset} but failed with DecodeError.",
                    )
                except LookupError:
                    yield InferenceTask().discard(
                        http_status=400,
                        err_msg=f"{self.__class__.__name__}: "
                        f"Unsupported charset {charset}",
                    )
        else:
            for d in input_:
                yield InferenceTask(inference_job_args=extra_args, data=d)

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[str]]
    ) -> ApiFuncArgs:
        return ([task.data for task in tasks],)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import platform
import json
import logging
import os
import uuid
from functools import lru_cache

from bentoml.utils.ruamel_yaml import YAML
from bentoml.utils import ProtoMessageToDict
from bentoml.configuration import _is_pip_installed_bentoml
from bentoml import __version__ as BENTOML_VERSION


logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def _amplitude_url():
    return "https://api.amplitude.com/httpapi"


@lru_cache(maxsize=1)
def _platform():
    return platform.platform()


@lru_cache(maxsize=1)
def _py_version():
    return "{major}.{minor}.{micro}".format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
        micro=sys.version_info.micro,
    )


@lru_cache(maxsize=1)
def _session_id():
    return str(uuid.uuid4())  # uuid that marks current python session


@lru_cache(maxsize=1)
def _api_key():
    if _is_pip_installed_bentoml():
        # Use prod amplitude key
        return '1ad6ee0e81b9666761aebd55955bbd3a'
    else:
        # Use dev amplitude key
        return '7f65f2446427226eb86f6adfacbbf47a'


def _send_amplitude_event(event_type, event_properties):
    """Send event to amplitude
    https://developers.amplitude.com/?objc--ios#http-api-v2-request-format
    """
    event = [
        {
            "event_type": event_type,
            "user_id": _session_id(),
            "event_properties": event_properties,
            "ip": "$remote",
        }
    ]
    event_data = {"api_key": _api_key(), "event": json.dumps(event)}

    try:
        import requests

        return requests.post(_amplitude_url(), data=event_data, timeout=1)
    except Exception as err:  # pylint:disable=broad-except
        # silently fail since this error does not concern BentoML end users
        logger.debug(str(err))


def _get_bento_service_event_properties(bento_service, properties=None):
    bento_service_metadata = bento_service.get_bento_service_metadata_pb()

    if properties is None:
        properties = {}

    artifact_types = set()
    input_types = set()
    output_types = set()

    for artifact in bento_service_metadata.artifacts:
        artifact_types.add(artifact.artifact_type)

    for api in bento_service_metadata.apis:
        input_types.add(api.input_type)
        output_types.add(api.output_type)

    if input_types:
        properties["input_types"] = list(input_types)
    if output_types:
        properties["output_types"] = list(output_types)

    if artifact_types:
        properties["artifact_types"] = list(artifact_types)
    else:
        properties["artifact_types"] = ["NO_ARTIFACT"]

    env_dict = ProtoMessageToDict(bento_service_metadata.env)
    if 'conda_env' in env_dict:
        env_dict['conda_env'] = YAML().load(env_dict['conda_env'])
    properties['env'] = env_dict

    return properties


@lru_cache(maxsize=1)
def _do_not_track() -> bool:
    """Returns if tracking is disabled from the environment variable

    Returns True if and only if the environment variable is defined and has value True.
    The function is cached for better performance.
    """
    return os.environ.get("BENTOML_DO_NOT_TRACK", str(False)).lower() == "true"


def track(event_type, event_properties=None):
    if _do_not_track():
        return  # Usage tracking disabled

    if event_properties is None:
        event_properties = {}

    event_properties['py_version'] = _py_version()
    event_properties["bento_version"] = BENTOML_VERSION
    event_properties["platform_info"] = _platform()

    return _send_amplitude_event(event_type, event_properties)


def track_save(bento_service, extra_properties=None):
    properties = _get_bento_service_event_properties(bento_service, extra_properties)
    return track("save", properties)



import logging
from typing import Sequence, Union

from bentoml.exceptions import MissingDependencyException

logger = logging.getLogger(__name__)


TF_KERAS_DEFAULT_FUNCTIONS = {
    "_default_save_signature",
    "call_and_return_all_conditional_losses",
}


TENSOR_CLASS_NAMES = (
    "RaggedTensor",
    "SparseTensor",
    "TensorArray",
    "EagerTensor",
    "Tensor",
)


def _isinstance(obj, klass: Union[str, type, Sequence]):
    if not klass:
        return False
    if isinstance(klass, str):
        return type(obj).__name__ == klass.split('.')[-1]
    if isinstance(klass, (tuple, list, set)):
        return any(_isinstance(obj, k) for k in klass)
    return isinstance(obj, klass)


def normalize_spec(value):
    if not _isinstance(value, TENSOR_CLASS_NAMES):
        return value

    import tensorflow as tf

    if _isinstance(value, "RaggedTensor"):
        return tf.RaggedTensorSpec.from_value(value)
    if _isinstance(value, "SparseTensor"):
        return tf.SparseTensorSpec.from_value(value)
    if _isinstance(value, "TensorArray"):
        return tf.TensorArraySpec.from_tensor(value)
    if _isinstance(value, ("Tensor", "EagerTensor")):
        return tf.TensorSpec.from_tensor(value)
    return value


def get_input_signatures(func):
    if hasattr(func, 'function_spec'):  # for RestoredFunction
        if func.function_spec.input_signature:
            return ((func.function_spec.input_signature, {}),)
        else:
            return tuple(
                s
                for conc in func.concrete_functions
                for s in get_input_signatures(conc)
            )
    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        if func.structured_input_signature is not None:
            return (func.structured_input_signature,)

        if func._arg_keywords is not None:  # TODO(bojiang): using private API
            return (
                (
                    tuple(),
                    {
                        k: normalize_spec(v)
                        for k, v in zip(func._arg_keywords, func.inputs)
                    },
                ),
            )
    return tuple()


def get_arg_names(func):
    if hasattr(func, 'function_spec'):  # for RestoredFunction
        return func.function_spec.arg_names
    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        return func._arg_keywords
    return tuple()


def get_output_signature(func):
    if hasattr(func, 'function_spec'):  # for RestoredFunction
        # assume all concrete functions have same signature
        return get_output_signature(func.concrete_functions[0])

    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        if func.structured_outputs is not None:
            if isinstance(func.structured_outputs, dict):
                return {
                    k: normalize_spec(v) for k, v in func.structured_outputs.items()
                }
            return func.structured_outputs
        else:
            return tuple(normalize_spec(v) for v in func.outputs)

    return tuple()


def get_restored_functions(m):
    function_map = {k: getattr(m, k, None) for k in dir(m)}
    return {
        k: v
        for k, v in function_map.items()
        if k not in TF_KERAS_DEFAULT_FUNCTIONS and hasattr(v, 'function_spec')
    }


def get_serving_default_function(m):
    try:
        import tensorflow as tf
    except ImportError:
        raise MissingDependencyException(
            "Tensorflow package is required to use TfSavedModelArtifact"
        )

    return m.signatures.get(tf.compat.v2.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY)


def cast_tensor_by_spec(_input, spec):
    '''
    transform dtype & shape following spec
    '''
    try:
        import tensorflow as tf
    except ImportError:
        raise MissingDependencyException(
            "Tensorflow package is required to use TfSavedModelArtifact"
        )

    if not _isinstance(spec, "TensorSpec"):
        return _input

    if _isinstance(_input, ["Tensor", "EagerTensor"]):
        # TensorFlow issue #43038
        # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
        return tf.cast(_input, dtype=spec.dtype, name=spec.name)
    else:
        return tf.constant(_input, dtype=spec.dtype, name=spec.name)


def _pretty_format_function_call(base, name, arg_names):
    if arg_names:
        part_sigs = ', '.join(f"{k}" for k in arg_names)
    else:
        part_sigs = ''

    if name == "__call__":
        return f"{base}({part_sigs})"
    return f"{base}.{name}({part_sigs})"


def _pretty_format_positional(positional):
    return "Positional arguments ({} total):\n    * {}".format(
        len(positional), "\n    * ".join(str(a) for a in positional)
    )


def pretty_format_function(function, obj="<object>", name="<function>"):
    ret = ''
    outs = get_output_signature(function)
    sigs = get_input_signatures(function)
    arg_names = get_arg_names(function)

    if hasattr(function, 'function_spec'):
        arg_names = function.function_spec.arg_names
    else:
        arg_names = function._arg_keywords

    ret += _pretty_format_function_call(obj, name, arg_names)
    ret += '\n------------\n'

    signature_descriptions = []

    for index, sig in enumerate(sigs):
        positional, keyword = sig
        signature_descriptions.append(
            "Arguments Option {}:\n  {}\n  Keyword arguments:\n    {}".format(
                index + 1, _pretty_format_positional(positional), keyword
            )
        )

    ret += '\n\n'.join(signature_descriptions)
    ret += f"\n\nReturn:\n  {outs}\n\n"
    return ret


def pretty_format_restored_model(model):
    part_functions = ""

    restored_functions = get_restored_functions(model)
    for name, func in restored_functions.items():
        part_functions += pretty_format_function(func, "model", name)
        part_functions += "\n"

    serving_default = get_serving_default_function(model)
    if serving_default:
        part_functions += pretty_format_function(
            serving_default, "model", "signature['serving_default']"
        )
        part_functions += "\n"

    if not restored_functions and not serving_default:
        return (
            "No serving function was found in the saved model. "
            "In the model implementation, use `tf.function` decorator to mark "
            "the method needed for model serving. \n"
            "Find more details in related TensorFlow docs here "
            "https://www.tensorflow.org/api_docs/python/tf/saved_model/save"
        )
    return f"Found restored functions:\n{part_functions}"



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer


@inject
def get_open_api_spec_json(
    bento_service,
    enable_metrics: bool = Provide[
        BentoMLContainer.config.bento_server.metrics.enabled
    ],
    enable_feedback: bool = Provide[
        BentoMLContainer.config.bento_server.feedback.enabled
    ],
):
    """
    The docs for all endpoints in Open API format.
    """
    docs = OrderedDict(
        openapi="3.0.0",
        info=OrderedDict(
            version=bento_service.version,
            title=bento_service.name,
            description="To get a client SDK, copy all content from <a "
            "href=\"/docs.json\">docs</a> and paste into "
            "<a href=\"https://editor.swagger.io\">editor.swagger.io</a> then click "
            "the tab <strong>Generate Client</strong> and choose the language.",
        ),
        tags=[{"name": "infra"}, {"name": "app"}],
    )

    paths = OrderedDict()
    default_response = {"200": {"description": "success"}}

    paths["/healthz"] = OrderedDict(
        get=OrderedDict(
            tags=["infra"],
            description="Health check endpoint. Expecting an empty response with status"
            " code 200 when the service is in health state",
            responses=default_response,
        )
    )

    paths["/metadata"] = OrderedDict(
        get=OrderedDict(
            tags=["infra"],
            description="BentoService metadata endpoint. Returns the service's"
            "`bentoml.yml` in JSON format.",
            responses=default_response,
        )
    )

    if enable_metrics:
        paths["/metrics"] = OrderedDict(
            get=OrderedDict(
                tags=["infra"],
                description="Prometheus metrics endpoint",
                responses=default_response,
            )
        )
    if enable_feedback:
        paths["/feedback"] = OrderedDict(
            post=OrderedDict(
                tags=["infra"],
                description="Provide feedback to prediction results from BentoService. "
                "Expecting feedback request payload in JSON format "
                "and requires `request_id` field, which can be obtained "
                "from any BentoService prediction response's header. "
                "Only last key will be considered if keys are repeated.",
                requestBody=OrderedDict(
                    required=True,
                    content={
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "required": ["request_id"],
                                "properties": {"request_id": {"type": "uuid"}},
                            },
                            "example": {
                                "request_id": "cf420b0f-15fa-013d-a37b-12345678c321",
                                "example_feedback": "key-value pair can be anything",
                            },
                        }
                    },
                ),
                responses=default_response,
            ),
        )

    for api in bento_service.inference_apis:
        path = "/{}".format(api.route)
        paths[path] = OrderedDict(
            post=OrderedDict(
                tags=["app"],
                description=api.doc,
                requestBody=OrderedDict(required=True, content=api.request_schema),
                responses=default_response,
            )
        )

    docs["paths"] = paths
    return docs



"""add last_updated_at for deployments

Revision ID: a6b00ae45279
Revises: 095fb029da39
Create Date: 2019-10-15 15:59:25.742280

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a6b00ae45279'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('deployments', sa.Column('last_updated_at', sa.DateTime))


def downgrade():
    op.drop_column('deployments', 'last_updated_at')



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import logging
import os
import sys

from flask import Flask, Response, jsonify, make_response, request, send_from_directory
from google.protobuf.json_format import MessageToJson
from simple_di import Provide, inject
from werkzeug.exceptions import BadRequest, NotFound

from bentoml.configuration import get_debug_mode
from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException
from bentoml.marshal.utils import DataLoader, MARSHAL_REQUEST_HEADER
from bentoml.server.instruments import InstrumentMiddleware
from bentoml.service import BentoService, InferenceAPI
from bentoml.tracing import get_tracer
from bentoml.types import HTTPRequest
from bentoml.utils.open_api import get_open_api_spec_json

CONTENT_TYPE_LATEST = str("text/plain; version=0.0.4; charset=utf-8")

feedback_logger = logging.getLogger("bentoml.feedback")
logger = logging.getLogger(__name__)


DEFAULT_INDEX_HTML = '''\
<!DOCTYPE html>
<head>
  <link rel="stylesheet" type="text/css" href="static_content/main.css">
  <link rel="stylesheet" type="text/css" href="static_content/readme.css">
  <link rel="stylesheet" type="text/css" href="static_content/swagger-ui.css">
</head>
<body>
  <div id="tab">
    <button
      class="tabLinks active"
      onclick="openTab(event, 'swagger_ui_container')"
      id="defaultOpen"
    >
      Swagger UI
    </button>
    <button class="tabLinks" onclick="openTab(event, 'markdown_readme')">
      ReadMe
    </button>
  </div>
  <script>
    function openTab(evt, tabName) {{
      // Declare all variables
      var i, tabContent, tabLinks;
      // Get all elements with class="tabContent" and hide them
      tabContent = document.getElementsByClassName("tabContent");
      for (i = 0; i < tabContent.length; i++) {{
        tabContent[i].style.display = "none";
      }}

      // Get all elements with class="tabLinks" and remove the class "active"
      tabLinks = document.getElementsByClassName("tabLinks");
      for (i = 0; i < tabLinks.length; i++) {{
        tabLinks[i].className = tabLinks[i].className.replace(" active", "");
      }}

      // Show the current tab, and add an "active" class to the button that opened the
      // tab
      document.getElementById(tabName).style.display = "block";
      evt.currentTarget.className += " active";
    }}
  </script>
  <div id="markdown_readme" class="tabContent"></div>
  <script src="static_content/marked.min.js"></script>
  <script>
    var markdownContent = marked(`{readme}`);
    var element = document.getElementById('markdown_readme');
    element.innerHTML = markdownContent;
  </script>
  <div id="swagger_ui_container" class="tabContent" style="display: block"></div>
  <script src="static_content/swagger-ui-bundle.js"></script>
  <script>
      SwaggerUIBundle({{
          url: '{url}',
          dom_id: '#swagger_ui_container'
      }})
  </script>
</body>
'''

SWAGGER_HTML = '''\
<!DOCTYPE html>
<head>
  <link rel="stylesheet" type="text/css" href="static_content/swagger-ui.css">
</head>
<body>
  <div id="swagger-ui-container"></div>
  <script src="static_content/swagger-ui-bundle.js"></script>
  <script>
      SwaggerUIBundle({{
          url: '{url}',
          dom_id: '#swagger-ui-container'
      }})
  </script>
</body>
'''


def _request_to_json(req):
    """
    Return request data for log prediction
    """
    if req.content_type == "application/json":
        return req.get_json()

    return {}


def log_exception(exc_info):
    """
    Logs an exception.  This is called by :meth:`handle_exception`
    if debugging is disabled and right before the handler is called.
    The default implementation logs the exception as error on the
    :attr:`logger`.
    """
    logger.error(
        "Exception on %s [%s]", request.path, request.method, exc_info=exc_info
    )


class BentoAPIServer:
    """
    BentoAPIServer creates a REST API server based on APIs defined with a BentoService
    via BentoService#get_service_apis call. Each InferenceAPI will become one
    endpoint exposed on the REST server, and the RequestHandler defined on each
    InferenceAPI object will be used to handle Request object before feeding the
    request data into a Service API function
    """

    @inject
    def __init__(
        self,
        bento_service: BentoService,
        app_name: str = None,
        enable_swagger: bool = Provide[
            BentoMLContainer.config.bento_server.swagger.enabled
        ],
        enable_metrics: bool = Provide[
            BentoMLContainer.config.bento_server.metrics.enabled
        ],
        enable_feedback: bool = Provide[
            BentoMLContainer.config.bento_server.feedback.enabled
        ],
    ):
        app_name = bento_service.name if app_name is None else app_name

        self.bento_service = bento_service
        self.app = Flask(app_name, static_folder=None)
        self.static_path = self.bento_service.get_web_static_content_path()
        self.enable_swagger = enable_swagger
        self.enable_metrics = enable_metrics
        self.enable_feedback = enable_feedback

        self.swagger_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'static_content'
        )

        for middleware in (InstrumentMiddleware,):
            self.app.wsgi_app = middleware(self.app.wsgi_app, self.bento_service)

        self.setup_routes()

    def start(self, port: int, host: str = "127.0.0.1"):
        """
        Start an REST server at the specific port on the instance or parameter.
        """
        # Bentoml api service is not thread safe.
        # Flask dev server enabled threaded by default, disable it.
        self.app.run(
            host=host,
            port=port,
            threaded=False,
            debug=get_debug_mode(),
            use_reloader=False,
        )

    @staticmethod
    def static_serve(static_path, file_path):
        """
        The static files route for BentoML API server
        """
        try:
            return send_from_directory(static_path, file_path)
        except NotFound:
            return send_from_directory(
                os.path.join(static_path, file_path), "index.html"
            )

    @staticmethod
    def index_view_func(static_path):
        """
        The index route for BentoML API server
        """
        return send_from_directory(static_path, 'index.html')

    def default_index_view_func(self):
        """
        The default index view for BentoML API server. This includes the readme
        generated from docstring and swagger UI
        """
        if not self.enable_swagger:
            return Response(
                response="Swagger is disabled", status=404, mimetype="text/html"
            )
        return Response(
            response=DEFAULT_INDEX_HTML.format(
                url='/docs.json', readme=self.bento_service.__doc__
            ),
            status=200,
            mimetype="text/html",
        )

    def swagger_ui_func(self):
        """
        The swagger UI route for BentoML API server
        """
        if not self.enable_swagger:
            return Response(
                response="Swagger is disabled", status=404, mimetype="text/html"
            )
        return Response(
            response=SWAGGER_HTML.format(url='/docs.json'),
            status=200,
            mimetype="text/html",
        )

    @staticmethod
    def swagger_static(static_path, filename):
        """
        The swagger static files route for BentoML API server
        """
        return send_from_directory(static_path, filename)

    @staticmethod
    def docs_view_func(bento_service):
        docs = get_open_api_spec_json(bento_service)
        return jsonify(docs)

    @staticmethod
    def healthz_view_func():
        """
        Health check for BentoML API server.
        Make sure it works with Kubernetes liveness probe
        """
        return Response(response="\n", status=200, mimetype="text/plain")

    @staticmethod
    def metadata_json_func(bento_service):
        bento_service_metadata = bento_service.get_bento_service_metadata_pb()
        return jsonify(MessageToJson(bento_service_metadata))

    def metrics_view_func(self):
        # noinspection PyProtectedMember
        from prometheus_client import generate_latest

        return generate_latest()

    @staticmethod
    def feedback_view_func(bento_service):
        """
        User send feedback along with the request_id. It will be stored is feedback logs
        ready for further process.
        """
        data = request.get_json()

        if not data:
            raise BadRequest("Failed parsing feedback JSON data")

        if "request_id" not in data:
            raise BadRequest("Missing 'request_id' in feedback JSON data")

        data["service_name"] = bento_service.name
        data["service_version"] = bento_service.version
        feedback_logger.info(data)
        return "success"

    def setup_routes(self):
        """
        Setup routes for bento model server, including:

        /               Index Page
        /docs           Swagger UI
        /healthz        Health check ping
        /feedback       Submitting feedback
        /metrics        Prometheus metrics endpoint
        /metadata       BentoService Artifact Metadata

        And user defined InferenceAPI list into flask routes, e.g.:
        /classify
        /predict
        """
        if self.static_path:
            # serve static files for any given path
            # this will also serve index.html from directory /any_path/
            # for path as /any_path/
            self.app.add_url_rule(
                "/<path:file_path>",
                "static_proxy",
                partial(self.static_serve, self.static_path),
            )
            # serve index.html from the directory /any_path
            # for path as /any_path/index
            self.app.add_url_rule(
                "/<path:file_path>/index",
                "static_proxy2",
                partial(self.static_serve, self.static_path),
            )
            # serve index.html from root directory for path as /
            self.app.add_url_rule(
                "/", "index", partial(self.index_view_func, self.static_path)
            )
        else:
            self.app.add_url_rule("/", "index", self.default_index_view_func)

        self.app.add_url_rule("/docs", "swagger", self.swagger_ui_func)
        self.app.add_url_rule(
            "/static_content/<path:filename>",
            "static_content",
            partial(self.swagger_static, self.swagger_path),
        )
        self.app.add_url_rule(
            "/docs.json", "docs", partial(self.docs_view_func, self.bento_service)
        )
        self.app.add_url_rule("/healthz", "healthz", self.healthz_view_func)
        self.app.add_url_rule(
            "/metadata",
            "metadata",
            partial(self.metadata_json_func, self.bento_service),
        )

        if self.enable_metrics:
            self.app.add_url_rule("/metrics", "metrics", self.metrics_view_func)

        if self.enable_feedback:
            self.app.add_url_rule(
                "/feedback",
                "feedback",
                partial(self.feedback_view_func, self.bento_service),
                methods=["POST"],
            )

        self.setup_bento_service_api_routes()

    def setup_bento_service_api_routes(self):
        """
        Setup a route for each InferenceAPI object defined in bento_service
        """
        for api in self.bento_service.inference_apis:
            route_function = self.bento_service_api_func_wrapper(api)
            self.app.add_url_rule(
                rule="/{}".format(api.route),
                endpoint=api.name,
                view_func=route_function,
                methods=api.input_adapter.HTTP_METHODS,
            )

    def bento_service_api_func_wrapper(self, api: InferenceAPI):
        """
        Create api function for flask route, it wraps around user defined API
        callback and adapter class, and adds request logging and instrument metrics
        """

        def api_func():
            # handle_request may raise 4xx or 5xx exception.
            try:
                if request.headers.get(MARSHAL_REQUEST_HEADER):
                    reqs = DataLoader.split_requests(request.get_data())
                    responses = api.handle_batch_request(reqs)
                    response_body = DataLoader.merge_responses(responses)
                    response = make_response(response_body)
                else:
                    req = HTTPRequest.from_flask_request(request)
                    resp = api.handle_request(req)
                    response = resp.to_flask_response()
            except BentoMLException as e:
                log_exception(sys.exc_info())

                if 400 <= e.status_code < 500 and e.status_code not in (401, 403):
                    response = make_response(
                        jsonify(
                            message="BentoService error handling API request: %s"
                            % str(e)
                        ),
                        e.status_code,
                    )
                else:
                    response = make_response('', e.status_code)
            except Exception:  # pylint: disable=broad-except
                # For all unexpected error, return 500 by default. For example,
                # if users' model raises an error of division by zero.
                log_exception(sys.exc_info())

                response = make_response(
                    'An error has occurred in BentoML user code when handling this '
                    'request, find the error details in server logs',
                    500,
                )

            return response

        def api_func_with_tracing():
            with get_tracer().span(
                service_name=f"BentoService.{self.bento_service.name}",
                span_name=f"InferenceAPI {api.name} HTTP route",
                request_headers=request.headers,
            ):
                return api_func()

        return api_func_with_tracing



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import glob
import gzip
import importlib
import io
import json
import logging
import os
import shutil
import stat
import tarfile
from urllib.parse import urlparse

from typing import TYPE_CHECKING
import requests

from bentoml.configuration import _is_pip_installed_bentoml
from bentoml.exceptions import BentoMLException
from bentoml.saved_bundle.local_py_modules import (
    copy_local_py_modules,
    copy_zip_import_archives,
)
from bentoml.saved_bundle.loader import _is_remote_path
from bentoml.saved_bundle.templates import (
    BENTO_SERVICE_BUNDLE_SETUP_PY_TEMPLATE,
    INIT_PY_TEMPLATE,
    MANIFEST_IN_TEMPLATE,
    MODEL_SERVER_DOCKERFILE_CPU,
)
from bentoml.utils import is_gcs_url, is_s3_url, archive_directory_to_tar
from bentoml.utils.tempdir import TempDirectory
from bentoml.utils.usage_stats import track_save
from bentoml.saved_bundle.config import SavedBundleConfig
from bentoml.saved_bundle.pip_pkg import get_zipmodules, ZIPIMPORT_DIR
from bentoml.utils.open_api import get_open_api_spec_json

if TYPE_CHECKING:
    from bentoml.service import BentoService

DEFAULT_SAVED_BUNDLE_README = """\
# Generated BentoService bundle - {}:{}

This is a ML Service bundle created with BentoML, it is not recommended to edit
code or files contained in this directory. Instead, edit the code that uses BentoML
to create this bundle, and save a new BentoService bundle.
"""

logger = logging.getLogger(__name__)


def _write_bento_content_to_dir(bento_service: "BentoService", path: str):
    if not os.path.exists(path):
        raise BentoMLException("Directory '{}' not found".format(path))

    for artifact in bento_service.artifacts.get_artifact_list():
        if not artifact.packed:
            logger.warning(
                "Missing declared artifact '%s' for BentoService '%s'",
                artifact.name,
                bento_service.name,
            )
    module_base_path = os.path.join(path, bento_service.name)
    try:
        os.mkdir(module_base_path)
    except FileExistsError:
        raise BentoMLException(
            f"Existing module file found for BentoService {bento_service.name}"
        )

    # write README.md with custom BentoService's docstring if presented
    saved_bundle_readme = DEFAULT_SAVED_BUNDLE_README.format(
        bento_service.name, bento_service.version
    )
    if bento_service.__class__.__doc__:
        saved_bundle_readme += "\n"
        saved_bundle_readme += bento_service.__class__.__doc__.strip()

    with open(os.path.join(path, "README.md"), "w") as f:
        f.write(saved_bundle_readme)

    # save all model artifacts to 'base_path/name/artifacts/' directory
    bento_service.artifacts.save(module_base_path)

    # write conda environment, requirement.txt
    bento_service.env.infer_pip_packages(bento_service)
    bento_service.env.save(path)

    # Copy all local python modules used by the module containing the `bento_service`'s
    # class definition to saved bundle directory
    module_name, module_file = copy_local_py_modules(
        bento_service.__class__.__module__, os.path.join(path, bento_service.name)
    )

    # create __init__.py
    with open(os.path.join(path, bento_service.name, "__init__.py"), "w") as f:
        f.write(
            INIT_PY_TEMPLATE.format(
                service_name=bento_service.name,
                module_name=module_name,
                pypi_package_version=bento_service.version,
            )
        )

    # write setup.py, this make saved BentoService bundle pip installable
    setup_py_content = BENTO_SERVICE_BUNDLE_SETUP_PY_TEMPLATE.format(
        name=bento_service.name,
        pypi_package_version=bento_service.version,
        long_description=saved_bundle_readme,
    )
    with open(os.path.join(path, "setup.py"), "w") as f:
        f.write(setup_py_content)

    with open(os.path.join(path, "MANIFEST.in"), "w") as f:
        f.write(MANIFEST_IN_TEMPLATE.format(service_name=bento_service.name))

    # write Dockerfile
    logger.debug("Using Docker Base Image %s", bento_service._env._docker_base_image)
    with open(os.path.join(path, "Dockerfile"), "w") as f:
        f.write(
            MODEL_SERVER_DOCKERFILE_CPU.format(
                docker_base_image=bento_service._env._docker_base_image
            )
        )

    # copy custom web_static_content if enabled
    if bento_service.web_static_content:
        src_web_static_content_dir = os.path.join(
            os.getcwd(), bento_service.web_static_content
        )
        if not os.path.isdir(src_web_static_content_dir):
            raise BentoMLException(
                f'web_static_content directory {src_web_static_content_dir} not found'
            )
        dest_web_static_content_dir = os.path.join(
            module_base_path, 'web_static_content'
        )
        shutil.copytree(src_web_static_content_dir, dest_web_static_content_dir)

    # Copy docker-entrypoint.sh
    docker_entrypoint_sh_file_src = os.path.join(
        os.path.dirname(__file__), "docker-entrypoint.sh"
    )
    docker_entrypoint_sh_file_dst = os.path.join(path, "docker-entrypoint.sh")
    shutil.copyfile(docker_entrypoint_sh_file_src, docker_entrypoint_sh_file_dst)
    # chmod +x docker-entrypoint.sh
    st = os.stat(docker_entrypoint_sh_file_dst)
    os.chmod(docker_entrypoint_sh_file_dst, st.st_mode | stat.S_IEXEC)

    # copy bentoml-init.sh for install targz bundles
    bentoml_init_sh_file_src = os.path.join(
        os.path.dirname(__file__), "bentoml-init.sh"
    )
    bentoml_init_sh_file_dst = os.path.join(path, "bentoml-init.sh")
    shutil.copyfile(bentoml_init_sh_file_src, bentoml_init_sh_file_dst)
    # chmod +x bentoml_init_script file
    st = os.stat(bentoml_init_sh_file_dst)
    os.chmod(bentoml_init_sh_file_dst, st.st_mode | stat.S_IEXEC)

    # write bentoml.yml
    config = SavedBundleConfig(bento_service)
    config["metadata"].update({"module_name": module_name, "module_file": module_file})

    config.write_to_path(path)
    # Also write bentoml.yml to module base path to make it accessible
    # as package data after pip installed as a python package
    config.write_to_path(module_base_path)

    bundled_pip_dependencies_path = os.path.join(path, 'bundled_pip_dependencies')
    _bundle_local_bentoml_if_installed_from_source(bundled_pip_dependencies_path)
    # delete mtime and sort file in tarballs to normalize the checksums
    for tarball_file_path in glob.glob(
        os.path.join(bundled_pip_dependencies_path, '*.tar.gz')
    ):
        normalize_gztarball(tarball_file_path)

    # write open-api-spec json file
    with open(os.path.join(path, "docs.json"), "w") as f:
        json.dump(get_open_api_spec_json(bento_service), f, indent=2)


def save_to_dir(
    bento_service: "BentoService", path: str, version: str = None, silent: bool = False
) -> None:
    """
    Save given :class:`~bentoml.BentoService` along with all its artifacts,
    source code and dependencies to target file path, assuming path
    exist and empty. If target path is not empty, this call may override
    existing files in the given path.

    Args:
        bento_service (:class:`~bentoml.service.BentoService`):
            a BentoService instance
        path (`str`):
            Destination of where the bento service will be saved. The
            destination can be local path or remote path. The remote
            path supports both AWS S3('s3://bucket/path') and
            Google Cloud Storage('gs://bucket/path').
        version (`str`, `optional`):
            Override the service version with given version string
        silent (`bool`, `optional`):
            whether to hide the log message showing target save path
    """
    track_save(bento_service)

    from bentoml.service import BentoService

    if not isinstance(bento_service, BentoService):
        raise BentoMLException(
            "save_to_dir only works with instances of custom BentoService class"
        )

    if version is not None:
        # If parameter version provided, set bento_service version
        # Otherwise it will bet set the first time the `version` property get accessed
        bento_service.set_version(version)

    if _is_remote_path(path):
        # If user provided path is an remote location, the bundle will first save to
        # a temporary directory and then upload to the remote location
        logger.info(
            'Saving bento to an remote path. BentoML will first save the bento '
            'to a local temporary directory and then upload to the remote path.'
        )
        with TempDirectory() as temp_dir:
            _write_bento_content_to_dir(bento_service, temp_dir)
            with TempDirectory() as tarfile_dir:
                tarfile_path, tarfile_name = archive_directory_to_tar(
                    temp_dir, tarfile_dir, bento_service.name
                )
                _upload_file_to_remote_path(path, tarfile_path, tarfile_name)
    else:
        _write_bento_content_to_dir(bento_service, path)

    copy_zip_import_archives(
        os.path.join(path, bento_service.name, ZIPIMPORT_DIR),
        bento_service.__class__.__module__,
        list(get_zipmodules().keys()),
        bento_service.env._zipimport_archives or [],
    )

    if not silent:
        logger.info(
            "BentoService bundle '%s:%s' created at: %s",
            bento_service.name,
            bento_service.version,
            path,
        )


def normalize_gztarball(file_path: str):
    MTIME = datetime.datetime(2000, 1, 1).timestamp()
    tar_io = io.BytesIO()

    with tarfile.open(file_path, "r:gz") as f:
        with tarfile.TarFile("bundle.tar", mode='w', fileobj=tar_io) as nf:
            names = sorted(f.getnames())
            for name in names:
                info = f.getmember(name)
                info.mtime = MTIME
                nf.addfile(info, f.extractfile(name))

    with open(file_path, "wb") as nf:
        with gzip.GzipFile("bundle.tar.gz", mode="w", fileobj=nf, mtime=MTIME) as gf:
            gf.write(tar_io.getvalue())


def _bundle_local_bentoml_if_installed_from_source(target_path):
    """
    if bentoml is installed in editor mode(pip install -e), this will build a source
    distribution with the local bentoml fork and add it to saved BentoService bundle
    path under bundled_pip_dependencies directory
    """

    # Find bentoml module path
    (module_location,) = importlib.util.find_spec('bentoml').submodule_search_locations

    bentoml_setup_py = os.path.abspath(os.path.join(module_location, '..', 'setup.py'))

    # this is for BentoML developer to create BentoService containing custom develop
    # branches of BentoML library, it is True only when BentoML module is installed in
    # development mode via "pip install --editable ."
    if not _is_pip_installed_bentoml() and os.path.isfile(bentoml_setup_py):
        logger.info(
            "Detected non-PyPI-released BentoML installed, copying local BentoML module"
            "files to target saved bundle path.."
        )

        # Create tmp directory inside bentoml module for storing the bundled
        # targz file. Since dist-dir can only be inside of the module directory
        bundle_dir_name = '__bentoml_tmp_sdist_build'
        source_dir = os.path.abspath(
            os.path.join(module_location, '..', bundle_dir_name)
        )

        if os.path.isdir(source_dir):
            shutil.rmtree(source_dir, ignore_errors=True)
        os.mkdir(source_dir)

        from setuptools import sandbox

        sandbox.run_setup(
            bentoml_setup_py,
            ['-q', 'sdist', '--format', 'gztar', '--dist-dir', bundle_dir_name],
        )

        # copy the generated targz to saved bundle directory and remove it from
        # bentoml module directory
        shutil.copytree(source_dir, target_path)

        # clean up sdist build files
        shutil.rmtree(source_dir)


def _upload_file_to_remote_path(remote_path, file_path: str, file_name: str):
    """Upload file to remote path
    """
    parsed_url = urlparse(remote_path)
    bucket_name = parsed_url.netloc
    object_prefix_path = parsed_url.path.lstrip('/')
    object_path = f'{object_prefix_path}/{file_name}'
    if is_s3_url(remote_path):
        try:
            import boto3
        except ImportError:
            raise BentoMLException(
                '"boto3" package is required for saving bento to AWS S3 bucket'
            )
        s3_client = boto3.client('s3')
        with open(file_path, 'rb') as f:
            s3_client.upload_fileobj(f, bucket_name, object_path)
    elif is_gcs_url(remote_path):
        try:
            from google.cloud import storage
        except ImportError:
            raise BentoMLException(
                '"google.cloud" package is required for saving bento to Google '
                'Cloud Storage'
            )
        gcs_client = storage.Client()
        bucket = gcs_client.bucket(bucket_name)
        blob = bucket.blob(object_path)
        blob.upload_from_filename(file_path)
    else:
        http_response = requests.put(remote_path)
        if http_response.status_code != 200:
            raise BentoMLException(
                f'Error uploading BentoService to {remote_path} '
                f'{http_response.status_code}'
            )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from bentoml.utils.lazy_loader import LazyLoader
from bentoml.cli.click_utils import (
    BentoMLCommandGroup,
    parse_bento_tag_callback,
    CLI_COLOR_SUCCESS,
    _echo,
    parse_labels_callback,
)
from bentoml.cli.deployment import _print_deployment_info, _print_deployments_info
from bentoml.cli.utils import Spinner
from bentoml.utils import get_default_yatai_client
from bentoml.yatai.deployment.azure_functions.constants import (
    AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS,
    AZURE_FUNCTIONS_AUTH_LEVELS,
    DEFAULT_MIN_INSTANCE_COUNT,
    DEFAULT_MAX_BURST,
    DEFAULT_PREMIUM_PLAN_SKU,
    DEFAULT_FUNCTION_AUTH_LEVEL,
)
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.exceptions import CLIException
from bentoml.utils import status_pb_to_error_code_and_message

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


def get_azure_functions_sub_command():
    # pylint: disable=unused-variable

    @click.group(
        name='azure-functions',
        help='Commands for Azure Functions BentoService deployment',
        cls=BentoMLCommandGroup,
    )
    def azure_functions():
        pass

    @azure_functions.command(help='Deploy BentoService to Azure Functions')
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.argument('name', type=click.STRING)
    @click.option(
        '-b',
        '--bento',
        '--bento-service-bundle',
        type=click.STRING,
        required=True,
        callback=parse_bento_tag_callback,
        help='Target BentoService to be deployed, referenced by its name and version '
        'in the format of name:version. For example: "iris_classifier:v1.2.0"',
    )
    @click.option(
        '--location',
        type=click.STRING,
        help='The Azure location name for the deployment',
        required=True,
    )
    @click.option(
        '--min-instances',
        type=click.INT,
        default=DEFAULT_MIN_INSTANCE_COUNT,
        help=f'The minimum number of workers for the deployment. The default value is '
        f'{DEFAULT_MIN_INSTANCE_COUNT}',
    )
    @click.option(
        '--max-burst',
        type=click.INT,
        default=DEFAULT_MAX_BURST,
        help=f'The maximum number of elastic workers for the deployment. '
        f'The default value is {DEFAULT_MAX_BURST}',
    )
    @click.option(
        '--premium-plan-sku',
        type=click.Choice(AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS),
        default=DEFAULT_PREMIUM_PLAN_SKU,
        help=f'The Azure Functions premium SKU for the deployment. The default value '
        f'is {DEFAULT_PREMIUM_PLAN_SKU}',
    )
    @click.option(
        '-l',
        '--labels',
        type=click.STRING,
        callback=parse_labels_callback,
        help='Key:value pairs that are attached to deployments and intended to be used '
        'to specify identifying attributes of the deployments that are meaningful to '
        'users. Multiple labels are separated with `,`',
    )
    @click.option(
        '--function-auth-level',
        type=click.Choice(AZURE_FUNCTIONS_AUTH_LEVELS),
        default=DEFAULT_FUNCTION_AUTH_LEVEL,
        help=f'The authorization level for the deployed Azure Functions. The default '
        f'value is {DEFAULT_FUNCTION_AUTH_LEVEL}',
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will exit without waiting until it has verified '
        'cloud resource allocation',
    )
    def deploy(
        namespace,
        name,
        bento,
        location,
        min_instances,
        max_burst,
        premium_plan_sku,
        labels,
        function_auth_level,
        output,
        wait,
    ):
        _echo(
            message='Azure Functions deployment functionalities are being migrated to '
            'a separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/azure-functions-deploy '
            'going forward.',
            color='yellow',
        )
        bento_name, bento_version = bento.split(':')
        yatai_client = get_default_yatai_client()
        with Spinner(f'Deploying {bento} to Azure Functions'):
            result = yatai_client.deployment.create_azure_functions_deployment(
                name=name,
                namespace=namespace,
                labels=labels,
                bento_name=bento_name,
                bento_version=bento_version,
                location=location,
                min_instances=min_instances,
                max_burst=max_burst,
                premium_plan_sku=premium_plan_sku,
                function_auth_level=function_auth_level,
                wait=wait,
            )
            if result.status.status_code != yatai_proto.status_pb2.Status.OK:
                error_code, error_message = status_pb_to_error_code_and_message(
                    result.status
                )
                raise CLIException(f'{error_code}:{error_message}')
            _echo(
                f'Successfully created Azure Functions deployment {name}',
                CLI_COLOR_SUCCESS,
            )
            _print_deployment_info(result.deployment, output)

    @azure_functions.command(help='Update existing Azure Functions deployment')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-b',
        '--bento',
        '--bento-service-bundle',
        type=click.STRING,
        callback=parse_bento_tag_callback,
        help='Target BentoService to be deployed, referenced by its name and version '
        'in the format of name:version. For example: "iris_classifier:v1.2.0"',
    )
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, the default value is "dev" '
        'which can be changed in BentoML configuration file',
    )
    @click.option(
        '--min-instances',
        type=click.INT,
        help='The minimum number of workers for the deployment.',
    )
    @click.option(
        '--max-burst',
        type=click.INT,
        help='The maximum number of elastic workers for the deployment.',
    )
    @click.option(
        '--premium-plan-sku',
        type=click.Choice(AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS),
        help='The Azure Functions premium SKU for the deployment.',
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will exit without waiting until it has verified '
        'cloud resource allocation',
    )
    def update(
        name, namespace, bento, min_instances, max_burst, premium_plan_sku, output, wait
    ):
        _echo(
            message='Azure Functions deployment functionalities are being migrated to '
            'a separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/azure-functions-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        if bento:
            bento_name, bento_version = bento.split(':')
        else:
            bento_name = None
            bento_version = None
        with Spinner(f'Updating Azure Functions deployment {name}'):
            result = yatai_client.deployment.update_azure_functions_deployment(
                namespace=namespace,
                deployment_name=name,
                bento_name=bento_name,
                bento_version=bento_version,
                min_instances=min_instances,
                max_burst=max_burst,
                premium_plan_sku=premium_plan_sku,
                wait=wait,
            )
            if result.status.status_code != yatai_proto.status_pb2.Status.OK:
                error_code, error_message = status_pb_to_error_code_and_message(
                    result.status
                )
                raise CLIException(f'{error_code}:{error_message}')
            _echo(
                f'Successfully updated Azure Functions deployment {name}',
                CLI_COLOR_SUCCESS,
            )
            _print_deployment_info(result.deployment, output)

    @azure_functions.command(help='Delete Azure Functions deployment')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '--force',
        is_flag=True,
        help='force delete the deployment record in database and '
        'ignore errors when deleting cloud resources',
    )
    def delete(name, namespace, force):
        _echo(
            message='Azure Functions deployment functionalities are being migrated to '
            'a separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/azure-functions-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        get_deployment_result = yatai_client.deployment.get(namespace, name)
        if get_deployment_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_deployment_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        result = yatai_client.deployment.delete(name, namespace, force)
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(
            f'Successfully deleted Azure Functions deployment "{name}"',
            CLI_COLOR_SUCCESS,
        )

    @azure_functions.command(help='Get Azure Functions deployment information')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '-o', '--output', type=click.Choice(['json', 'yaml', 'table']), default='json'
    )
    def get(name, namespace, output):
        _echo(
            message='Azure Functions deployment functionalities are being migrated to '
            'a separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/azure-functions-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        get_result = yatai_client.deployment.get(namespace, name)
        if get_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        describe_result = yatai_client.deployment.describe(namespace, name)
        if describe_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        get_result.deployment.state.CopyFrom(describe_result.state)
        _print_deployment_info(get_result.deployment, output)

    @azure_functions.command(name='list', help='List Azure Functions deployments')
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
        default=ALL_NAMESPACE_TAG,
    )
    @click.option(
        '--limit',
        type=click.INT,
        help='The maximum amount of Azure Functions deployments to be listed at once',
    )
    @click.option(
        '-l',
        '--labels',
        type=click.STRING,
        help="Label query to filter Azure Functions deployments, supports '=', '!=', "
        "'IN', 'NotIn', 'Exists', and 'DoesNotExist'. (e.g. key1=value1, "
        "key2!=value2, key3 In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--order-by', type=click.Choice(['created_at', 'name']), default='created_at',
    )
    @click.option(
        '--asc/--desc',
        default=False,
        help='Ascending or descending order for list deployments',
    )
    @click.option(
        '-o', '--output', type=click.Choice(['json', 'yaml', 'table']), default='table'
    )
    def list_deployment(namespace, limit, labels, order_by, asc, output):
        _echo(
            message='Azure Functions deployment functionalities are being migrated to '
            'a separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/azure-functions-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        list_result = yatai_client.deployment.list_azure_functions_deployments(
            limit=limit,
            labels=labels,
            namespace=namespace,
            order_by=order_by,
            ascending_order=asc,
        )
        if list_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                list_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _print_deployments_info(list_result.deployments, output)

    return azure_functions



import atexit
from multiprocessing.util import _exit_function


def worker_exit(server, worker):  # pylint: disable=unused-argument
    from prometheus_client import multiprocess

    multiprocess.mark_process_dead(worker.pid)


def post_fork(server, worker):
    server.log.debug("Worker spawned (pid: %s)", worker.pid)


def pre_fork(server, worker):  # pylint: disable=unused-argument
    pass


def pre_exec(server):
    server.log.debug("Forked child, re-executing.")


def when_ready(server):
    server.log.debug("Server is ready. Spawning workers")


def worker_int(worker):
    worker.log.debug("worker received INT or QUIT signal")

    # get traceback info
    import threading
    import sys
    import traceback

    id2name = {th.ident: th.name for th in threading.enumerate()}
    code = []
    for threadId, stack in sys._current_frames().items():
        code.append("\n# Thread: %s(%d)" % (id2name.get(threadId, ""), threadId))
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                code.append("  %s" % (line.strip()))
    worker.log.debug("\n".join(code))


def worker_abort(worker):
    worker.log.debug("worker received SIGABRT signal")


def post_worker_init(worker):
    worker.log.debug('Unregistering usage tracking in worker process')
    atexit.unregister(_exit_function)  # Shutting down Gunicorn gracefully



import importlib
import os

from bentoml.exceptions import (
    ArtifactLoadingException,
    InvalidArgument,
    MissingDependencyException,
)
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv
from bentoml.utils import cloudpickle

MODULE_NAME_FILE_ENCODING = "utf-8"
KERAS_MODEL_EXTENSION = ".h5"


class KerasModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading Keras Model

    Args:
        name (string): name of the artifact
        default_custom_objects (dict): dictionary of Keras custom objects for model
        store_as_json_and_weights (bool): flag allowing storage of the Keras
            model as JSON and weights

    Raises:
        MissingDependencyException: keras or tensorflow.keras package is required for
            KerasModelArtifact
        InvalidArgument:  invalid argument type, model being packed must be instance of
            keras.engine.network.Network, tf.keras.models.Model, or their aliases

    Example usage:

    >>> from tensorflow import keras
    >>> from tensorflow.keras.models import Sequential
    >>> from tensorflow.keras.preprocessing import sequence, text
    >>>
    >>> model_to_save = Sequential()
    >>> # training model
    >>> model_to_save.compile(...)
    >>> model_to_save.fit(...)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.keras import KerasModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([KerasModelArtifact('model')])
    >>> class KerasModelService(bentoml.BentoService):
    >>>     @bentoml.api(input=JsonInput(), batch=False)
    >>>     def predict(self, parsed_json):
    >>>         input_data = text.text_to_word_sequence(parsed_json['text'])
    >>>         return self.artifacts.model.predict_classes(input_data)
    >>>
    >>> svc = KerasModelService()
    >>> svc.pack('model', model_to_save)
    """

    def __init__(
        self, name, default_custom_objects=None, store_as_json_and_weights=False,
    ):
        super(KerasModelArtifact, self).__init__(name)

        try:
            import tensorflow as tf
        except ImportError:
            raise MissingDependencyException(
                "Tensorflow package is required to use KerasModelArtifact. BentoML "
                "currently only support using Keras with Tensorflow backend."
            )

        self._store_as_json_and_weights = store_as_json_and_weights

        # By default assume using tf.keras module
        self._keras_module_name = tf.keras.__name__

        self._default_custom_objects = default_custom_objects
        self.graph = None
        self.sess = None

        self._model = None
        self._custom_objects = None

    def set_dependencies(self, env: BentoServiceEnv):
        # Note that keras module is not required, user can use tf.keras as an
        # replacement for the keras module. Although tensorflow module is required to
        #  be used as the default Keras backend
        if env._infer_pip_packages:
            pip_deps = ['tensorflow']
            if self._keras_module_name == 'keras':
                pip_deps.append('keras')
            env.add_pip_packages(pip_deps)

    def _keras_module_name_path(self, base_path):
        # The name of the keras module used, can be 'keras' or 'tensorflow.keras'
        return os.path.join(base_path, self.name + '_keras_module_name.txt')

    def _custom_objects_path(self, base_path):
        return os.path.join(base_path, self.name + '_custom_objects.pkl')

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + KERAS_MODEL_EXTENSION)

    def _model_weights_path(self, base_path):
        return os.path.join(base_path, self.name + '_weights.hdf5')

    def _model_json_path(self, base_path):
        return os.path.join(base_path, self.name + '_json.json')

    def bind_keras_backend_session(self):
        try:
            import tensorflow as tf
        except ImportError:
            raise MissingDependencyException(
                "Tensorflow package is required to use KerasModelArtifact. BentoML "
                "currently only support using Keras with Tensorflow backend."
            )

        self.sess = tf.compat.v1.keras.backend.get_session()
        self.graph = self.sess.graph

    def create_session(self):
        try:
            import tensorflow as tf
        except ImportError:
            raise MissingDependencyException(
                "Tensorflow package is required to use KerasModelArtifact. BentoML "
                "currently only support using Keras with Tensorflow backend."
            )

        self.graph = tf.compat.v1.get_default_graph()
        self.sess = tf.compat.v1.Session(graph=self.graph)
        tf.compat.v1.keras.backend.set_session(self.sess)

    def pack(self, data, metadata=None):  # pylint:disable=arguments-renamed
        try:
            import tensorflow as tf
        except ImportError:
            raise MissingDependencyException(
                "Tensorflow package is required to use KerasModelArtifact. BentoML "
                "currently only support using Keras with Tensorflow backend."
            )

        if isinstance(data, dict):
            model = data['model']
            custom_objects = (
                data['custom_objects']
                if 'custom_objects' in data
                else self._default_custom_objects
            )
        else:
            model = data
            custom_objects = self._default_custom_objects

        if not isinstance(model, tf.keras.models.Model):
            error_msg = (
                "KerasModelArtifact#pack expects model argument to be type: "
                "keras.engine.network.Network, tf.keras.models.Model, or their "
                "aliases, instead got type: {}".format(type(model))
            )
            try:
                import keras

                if not isinstance(model, keras.engine.network.Network):
                    raise InvalidArgument(error_msg)
                else:
                    self._keras_module_name = keras.__name__
            except ImportError:
                raise InvalidArgument(error_msg)

        self.bind_keras_backend_session()
        self._model = model
        self._custom_objects = custom_objects
        return self

    def load(self, path):
        if os.path.isfile(self._keras_module_name_path(path)):
            with open(self._keras_module_name_path(path), "rb") as text_file:
                keras_module_name = text_file.read().decode(MODULE_NAME_FILE_ENCODING)
                try:
                    keras_module = importlib.import_module(keras_module_name)
                except ImportError:
                    raise ArtifactLoadingException(
                        "Failed to import '{}' module when loading saved "
                        "KerasModelArtifact".format(keras_module_name)
                    )
        else:
            raise ArtifactLoadingException(
                "Failed to read keras model name from '{}' when loading saved "
                "KerasModelArtifact".format(self._keras_module_name_path(path))
            )

        if self._default_custom_objects is None and os.path.isfile(
            self._custom_objects_path(path)
        ):
            self._default_custom_objects = cloudpickle.load(
                open(self._custom_objects_path(path), 'rb')
            )

        if self._store_as_json_and_weights:
            # load keras model via json and weights if store_as_json_and_weights=True
            self.create_session()
            with self.graph.as_default():
                with self.sess.as_default():
                    with open(self._model_json_path(path), 'r') as json_file:
                        model_json = json_file.read()
                    model = keras_module.models.model_from_json(
                        model_json, custom_objects=self._default_custom_objects
                    )
                    model.load_weights(self._model_weights_path(path))
        else:
            # otherwise, load keras model via standard load_model
            model = keras_module.models.load_model(
                self._model_file_path(path),
                custom_objects=self._default_custom_objects,
            )
        return self.pack(model)

    def save(self, dst):
        # save the keras module name to be used when loading
        with open(self._keras_module_name_path(dst), "wb") as text_file:
            text_file.write(self._keras_module_name.encode(MODULE_NAME_FILE_ENCODING))

        # save custom_objects for model
        cloudpickle.dump(
            self._custom_objects, open(self._custom_objects_path(dst), "wb")
        )

        # save keras model using json and weights if requested
        if self._store_as_json_and_weights:
            with open(self._model_json_path(dst), "w") as json_file:
                json_file.write(self._model.to_json())
            self._model.save_weights(self._model_weights_path(dst))

        # otherwise, save standard keras model
        else:
            self._model.save(self._model_file_path(dst))

    def get(self):
        return self._model



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import traceback
from typing import BinaryIO, Iterable, Sequence, Tuple

from bentoml.adapters.multi_file_input import MultiFileInput
from bentoml.adapters.utils import (
    check_file_extension,
    get_default_accept_image_formats,
)
from bentoml.types import InferenceTask, JsonSerializable, Optional
from bentoml.utils.lazy_loader import LazyLoader

# BentoML optional dependencies, using lazy load to avoid ImportError
imageio = LazyLoader('imageio', globals(), 'imageio')
numpy = LazyLoader('numpy', globals(), 'numpy')


ApiFuncArgs = Tuple[Sequence[numpy.ndarray], Sequence[Optional[JsonSerializable]]]
AnnoImgTask = InferenceTask[Tuple[BinaryIO, BinaryIO]]  # image file bytes, json bytes


class AnnotatedImageInput(MultiFileInput):
    """Transform incoming image data from http request, cli or lambda event into a
    imageio array (a subclass of numpy.ndarray that has a meta attribute), while
    allowing an optional JSON file for image annotations (such as object bounding
    boxes, class labels, etc.)

    Transforms input image file into a numpy array, and loads JSON file as
    a JSON serializable Python object, providing them to user-defined
    API functions.

    Parameters
    ----------
    image_input_name : str
        An acceptable input name for HTTP request
        and function keyword argument. Default value is "image"

    annotation_input_name : str
        An acceptable input name for HTTP request
        and function keyword argument. Default value is "annotations"

    accept_image_formats : List[str]
        A list of acceptable image formats.
        Default value is loaded from bentoml config
        'apiserver/default_image_input_accept_file_extensions', which is
        set to ['.jpg', '.png', '.jpeg', '.tiff', '.webp', '.bmp'] by default.
        List of all supported format can be found here:
        https://imageio.readthedocs.io/en/stable/formats.html

    pilmode : str
        The pilmode to be used for reading image file into numpy
        array. Default value is 'RGB'.  Find more information at:
        https://imageio.readthedocs.io/en/stable/format_png-pil.html

    Raises
    ----------
    ImportError: imageio package is required to use AnnotatedImageInput

    Examples
    ----------

    .. code-block:: python

        from typing import Sequence

        from bentoml import BentoService, api, artifacts
        from bentoml.frameworks.tensorflow import TensorflowSavedModelArtifact
        from bentoml.adapters import AnnotatedImageInput

        CLASS_NAMES = ['cat', 'dog']

        @artifacts([TensorflowSavedModelArtifact('classifier')])
        class PetClassification(BentoService):
           @api(input=AnnotatedImageInput(), batch=True)
           def predict(
                   self,
                   image_list: 'Sequence[imageio.core.utils.Array]',
                   annotations_list: 'Sequence[JsonSerializable]',
               ) -> Sequence[str]:
               cropped_pets = some_pet_finder(image_list, annotations_list)
               results = self.artifacts.classifer.predict(cropped_pets)
               return [CLASS_NAMES[r] for r in results]

    Query with HTTP request performed by cURL::

        curl -i \\
          -F image=@test.jpg \\
          -F annotations=@test.json \\
          localhost:5000/predict

    OR by an HTML form that sends multipart data:

    .. code-block:: html

        <form action="http://localhost:8000" method="POST"
              enctype="multipart/form-data">
            <input name="image" type="file">
            <input name="annotations" type="file">
            <input type="submit">
        </form>

    OR by python requests:

    .. code-block:: python

        import requests

        with open("test.jpg", "rb") as f:
            image_bytes = f.read()
        with open("anno.json", "rb") as f:
            anno_bytes = f.read()

        files = {
            "image": ("test.jpg", image_bytes),
            "annotations": ("test.json", anno_bytes),
        }
        response = requests.post(your_url, files=files)

    .. code-block:: python

        import requests
        import PIL

        pil_image = PIL.Image.open('test.jpg')
        annotations = { "age": 10, "bar": "foo" }

        image_bytes = pil_image.tobytes()
        anno_bytes = json.dumps(annotations).encode('utf-8')

        files = {
            "image": ("test.jpg", image_bytes),
            "annotations": ("test.json", anno_bytes),
        }
        response = requests.post(your_url, files=files)

    Query with CLI command::

        bentoml run PyTorchFashionClassifier:latest predict \\
          --input-file-image test.jpg \\
          --input-file-annotations test.json

    OR infer all file pairs under a folder with ten pairs each batch::

        bentoml run PyTorchFashionClassifier:latest predict --max-batch-size 10 \\
          --input-file-image folder/*.jpg \\
          --input-file-annotations folder/*.json

    Note: jpg files and json files should be in same prefix like this::

        folder:
            - apple.jpg
            - apple.json
            - banana.jpg
            - banana.json
            ...
    """

    def __init__(
        self,
        image_input_name="image",
        annotation_input_name="annotations",
        pilmode="RGB",
        accept_image_formats=None,
        **base_kwargs,
    ):
        assert imageio, "`imageio` dependency can be imported"
        input_names = [image_input_name, annotation_input_name]
        super().__init__(input_names=input_names, allow_none=True, **base_kwargs)

        self.image_input_name = image_input_name
        self.annotation_input_name = annotation_input_name
        self.pilmode = pilmode
        self.accept_image_formats = (
            accept_image_formats or get_default_accept_image_formats()
        )

    @property
    def pip_dependencies(self):
        return ["imageio"]

    @property
    def config(self):
        return {
            # Converting to list, google.protobuf.Struct does not work with tuple type
            "accept_image_formats": list(self.accept_image_formats),
            "pilmode": self.pilmode,
        }

    @property
    def request_schema(self):
        return {
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {
                        self.image_input_name: {"type": "string", "format": "binary"},
                        self.annotation_input_name: {
                            "type": "string",
                            "format": "binary",
                        },
                    },
                }
            },
        }

    def extract_user_func_args(self, tasks: Iterable[AnnoImgTask]) -> ApiFuncArgs:
        image_arrays = []
        json_objs = []
        for task in tasks:
            try:
                image_file, json_file = task.data
                assert image_file is not None
                assert getattr(image_file, "name", None)
                assert check_file_extension(image_file.name, self.accept_image_formats)
                image_array = imageio.imread(image_file, pilmode=self.pilmode)
                image_arrays.append(image_array)
                if json_file is not None:
                    json_objs.append(json.load(json_file))
                else:
                    json_objs.append(None)
            except AssertionError:
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"Input image file must be in supported format list: "
                    f"{self.accept_image_formats}",
                )
            except UnicodeDecodeError:
                task.discard(
                    http_status=400, err_msg="JSON must be in unicode",
                )
            except json.JSONDecodeError:
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"received invalid JSON file",
                )
            except Exception:  # pylint: disable=broad-except
                err = traceback.format_exc()
                task.discard(
                    http_status=500,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"Internal Server Error: {err}",
                )
        return tuple(image_arrays), tuple(json_objs)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bentoml.exceptions import InvalidArgument


def raise_if_api_names_not_found_in_bento_service_metadata(metadata, api_names):
    all_api_names = [api.name for api in metadata.apis]

    if not set(api_names).issubset(all_api_names):
        raise InvalidArgument(
            "Expect api names {api_names} to be "
            "subset of {all_api_names}".format(
                api_names=api_names, all_api_names=all_api_names
            )
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
import sys
import tarfile
import logging

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)


def download_extra_resources():
    s3_bucket = os.environ.get('BENTOML_S3_BUCKET')
    s3_prefix = os.environ.get('BENTOML_DEPLOYMENT_PATH_PREFIX')
    additional_pkg_dir = '/tmp/requirements'

    if not os.path.exists(additional_pkg_dir) or not os.listdir(additional_pkg_dir):
        # Using print instead of logger.info, because this ran before bentoml is loaded.
        print('Additional required modules are not present. Downloading from s3')
        s3_file_path = os.path.join(s3_prefix, 'requirements.tar')

        s3_client = boto3.client('s3')
        print(f'requirement.tar does not exist, downloading from {s3_bucket}')
        try:
            obj = s3_client.get_object(Bucket=s3_bucket, Key=s3_file_path)
            bytestream = io.BytesIO(obj['Body'].read())
            print('Extracting required modules from tar file')
            tar = tarfile.open(fileobj=bytestream, mode='r:*')
            tar.extractall(path='/tmp')
            print('Appending /tmp/requirements to PYTHONPATH')
            sys.path.append(additional_pkg_dir)
        except ClientError as e:
            if e.response['Error']['Code'] == "404":
                print(f'File {s3_bucket}/{s3_file_path} does not exists')
                logger.error('File %s/%s does not exists', s3_bucket, s3_file_path)
            else:
                raise
    else:
        print('Additional required modules already present, skipping download from s3')



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask, Response, request

from bentoml.types import HTTPRequest


def setup_bento_service_api_route(app, api):
    def view_function():
        req = HTTPRequest.from_flask_request(request)
        response = api.handle_request(req)
        return response.to_flask_response()

    app.add_url_rule(
        rule="/invocations",
        endpoint=api.name,
        view_func=view_function,
        methods=["POST"],
    )


def ping_view_func():
    return Response(response="\n", status=200, mimetype="application/json")


def setup_routes(app, bento_service, api_name):
    """
    Setup routes required for AWS sagemaker
    /ping
    /invocations
    """
    app.add_url_rule("/ping", "ping", ping_view_func)
    api = bento_service.get_inference_api(api_name)
    setup_bento_service_api_route(app, api)


# AWS Sagemaker requires custom inference docker container to implement a web server
# that responds to /invocations and /ping on port 8080.
AWS_SAGEMAKER_SERVE_PORT = 8080


class BentomlSagemakerServer:
    """
    BentomlSagemakerServer create an AWS Sagemaker compatibility REST API model server
    """

    def __init__(self, bento_service, api_name, app_name=None):
        app_name = bento_service.name if app_name is None else app_name

        self.bento_service = bento_service
        self.app = Flask(app_name)
        setup_routes(self.app, self.bento_service, api_name)

    def start(self):
        self.app.run(port=AWS_SAGEMAKER_SERVE_PORT)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import functools
import itertools
import pathlib
import sys
from typing import Iterable, Iterator, NamedTuple, Sequence, Tuple

from bentoml.types import (
    ApiFuncArgs,
    AwsLambdaEvent,
    BatchApiFuncArgs,
    FileLike,
    HTTPRequest,
    InferenceTask,
)


class BaseInputAdapter:
    """
    InputAdapter is an abstraction layer between user defined API callback function
    and prediction request input in a variety of different forms, such as HTTP request
    body, command line arguments or AWS Lambda event object.
    """

    HTTP_METHODS = ["POST", "GET"]
    BATCH_MODE_SUPPORTED = True
    SINGLE_MODE_SUPPORTED = True

    def __init__(self, http_input_example=None, **base_config):
        self._config = base_config
        self._http_input_example = http_input_example
        self.custom_request_schema = base_config.get('request_schema')

    @property
    def config(self):
        return self._config

    @property
    def request_schema(self):
        """
        :return: OpenAPI json schema for the HTTP API endpoint created with this input
                 adapter
        """
        return {"application/json": {"schema": {"type": "object"}}}

    @property
    def pip_dependencies(self):
        """
        :return: List of PyPI package names required by this InputAdapter
        """
        return []

    def from_http_request(self, req: HTTPRequest) -> InferenceTask:
        """
        Handles HTTP requests, convert it into InferenceTask
        """
        raise NotImplementedError()

    def from_aws_lambda_event(self, event: AwsLambdaEvent) -> InferenceTask:
        """
        Handles AWS lambda events, convert it into InferenceTask
        """
        raise NotImplementedError()

    def from_cli(self, cli_args: Tuple[str, ...]) -> Iterator[InferenceTask]:
        """
        Handles CLI command, generate InferenceTask
        """
        raise NotImplementedError()

    def from_inference_job(self, *_args, **_kwargs) -> Iterator[InferenceTask]:
        """
        Handles <bento_inference_api>.run(), generate InferenceTask
        """
        raise NotImplementedError()

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask]
    ) -> BatchApiFuncArgs:
        """
        Extract args that user API function is expecting from InferenceTask
        """
        raise NotImplementedError()

    def iter_batch_args(
        self,
        batch_args: BatchApiFuncArgs,
        tasks: InferenceTask = None,  # pylint: disable=unused-argument
    ) -> Iterator[ApiFuncArgs]:
        """
        Extract args that user API function is expecting from InferenceTask
        """
        return iter(zip(*batch_args))


COLOR_FAIL = '\033[91m'


def exit_cli(err_msg: str = "", exit_code: int = None):
    if exit_code is None:
        exit_code = 1 if err_msg else 0
    if err_msg:
        print(f"{COLOR_FAIL}{err_msg}", file=sys.stderr)
    sys.exit(exit_code)


class CliInputParser(NamedTuple):
    arg_names: Tuple[str]
    file_arg_names: Tuple[str]
    arg_strs: Tuple[str]
    file_arg_strs: Tuple[str]
    parser: argparse.ArgumentParser

    @classmethod
    @functools.lru_cache()
    def get(cls, input_names: Tuple[str] = None):
        arg_names = (
            tuple(f"input_{n}" for n in input_names) if input_names else ("input",)
        )
        arg_strs = tuple(f'--{n.replace("_", "-")}' for n in arg_names)

        file_arg_names = (
            tuple(f"input_file_{n}" for n in input_names)
            if input_names
            else ("input_file",)
        )
        file_arg_strs = tuple(f'--{n.replace("_", "-")}' for n in file_arg_names)

        parser = argparse.ArgumentParser()
        for name in itertools.chain(arg_strs, file_arg_strs):
            parser.add_argument(name, nargs="+")

        return cls(arg_names, file_arg_names, arg_strs, file_arg_strs, parser)

    def parse(self, args: Sequence[str]):
        try:
            parsed, _ = self.parser.parse_known_args(args)
        except SystemExit:
            parsed = None

        inputs = tuple(getattr(parsed, name, None) for name in self.arg_names)
        file_inputs = tuple(getattr(parsed, name, None) for name in self.file_arg_names)

        if any(inputs) and any(file_inputs):
            exit_cli(
                '''
                Conflict arguments:
                --input* and --input-file* should not be provided at same time
                '''
            )
        if not all(inputs) and not all(file_inputs):
            exit_cli(
                f'''
                Insufficient arguments:
                ({' '.join(self.arg_strs)}) or
                ({' '.join(self.file_arg_strs)})
                are required
                '''
            )

        if all(inputs):
            if functools.reduce(lambda i, j: len(i) == len(j), inputs):
                return inputs, None
            else:
                exit_cli(
                    f'''
                    Arguments length mismatch:
                    Each ({' '.join(self.arg_strs)})
                    should have same amount of inputs
                    '''
                )

        if all(file_inputs):
            if functools.reduce(lambda i, j: len(i) == len(j), file_inputs):
                return None, file_inputs
            else:
                exit_cli(
                    f'''
                    Arguments length mismatch:
                    Each ({' '.join(self.file_arg_strs)})
                    should have same amount of inputs
                    '''
                )


def parse_cli_inputs(
    args: Sequence[str], input_names: Sequence[str] = None
) -> Iterator[Tuple[FileLike]]:
    '''
    Parse CLI args and iter each pair of inputs in bytes.

    >>> parse_cli_inputs("--input-x '1' '2' --input-y 'a' 'b'".split(' '), ('x', 'y'))
    >>> parse_cli_inputs(
    >>>     "--input-file-x 1.jpg 2.jpg --input-file-y 1.label 2.label".split(' '),
    >>>     ('x', 'y'))
    '''
    parser = CliInputParser.get(tuple(input_names))
    return parser.parse(args)


def parse_cli_input(cli_args: Iterable[str]) -> Iterator[FileLike]:
    '''
    Parse CLI args and iter each input in bytes.

    >>> parse_cli_input('--input {"input":1} {"input":2}'.split(' '))
    OR
    >>> parse_cli_inputs("--input-file 1.jpg 2.jpg 3.jpg".split(' '))
    '''
    parser = argparse.ArgumentParser()
    input_g = parser.add_mutually_exclusive_group(required=True)
    input_g.add_argument('--input', nargs="+", type=str)
    input_g.add_argument('--input-file', nargs="+")

    parsed_args, _ = parser.parse_known_args(list(cli_args))

    inputs = tuple(
        parsed_args.input if parsed_args.input_file is None else parsed_args.input_file
    )
    is_file = parsed_args.input_file is not None
    if is_file:
        for input_ in inputs:
            uri = pathlib.Path(input_).absolute().as_uri()
            yield FileLike(uri=uri)

    else:
        for input_ in inputs:
            rv = FileLike(bytes_=input_.encode())
            yield rv

    return _



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import BinaryIO, Iterable, Sequence, Tuple

from bentoml.adapters.file_input import FileInput
from bentoml.adapters.utils import (
    check_file_extension,
    get_default_accept_image_formats,
)
from bentoml.types import InferenceTask
from bentoml.utils.lazy_loader import LazyLoader

# BentoML optional dependencies, using lazy load to avoid ImportError
imageio = LazyLoader('imageio', globals(), 'imageio')
numpy = LazyLoader('numpy', globals(), 'numpy')


ApiFuncArgs = Tuple[
    Sequence['numpy.ndarray'],
]


class ImageInput(FileInput):
    """Convert incoming image data from http request, cli or lambda event into imageio
    array (a subclass of numpy.ndarray that has a meta attribute) and pass down to
    user defined API functions.

    ** To operate raw files or PIL.Image obj, use the low-level :class:`.FileInput`. **

    Parameters
    ----------
    accept_image_formats : List[str]
        A list of acceptable image formats.
        Default value is loaded from bentoml config
        'apiserver/default_image_input_accept_file_extensions', which is
        set to ['.jpg', '.png', '.jpeg', '.tiff', '.webp', '.bmp'] by default.
        List of all supported format can be found here:
        https://imageio.readthedocs.io/en/stable/formats.html
    pilmode : str
        The pilmode to be used for reading image file into numpy
        array. Default value is 'RGB'.  Find more information at:
        https://imageio.readthedocs.io/en/stable/format_png-pil.html

    Raises
    ----------
    ImportError: imageio package is required to use ImageInput

    Examples
    ----------

    Service using ImageInput:

    .. code-block:: python

        from typing import List

        import numpy as np
        from bentoml import BentoService, api, artifacts
        from bentoml.frameworks.tensorflow import TensorflowSavedModelArtifact
        from bentoml.adapters import ImageInput

        CLASS_NAMES = ['cat', 'dog']

        @artifacts([TensorflowSavedModelArtifact('classifier')])
        class PetClassification(BentoService):
            @api(input=ImageInput(), batch=True)
            def predict(
                self, image_arrays: List[imageio.core.utils.Array]
            ) -> List[str]:
                results = self.artifacts.classifer.predict(image_arrays)
                return [CLASS_NAMES[r] for r in results]

    OR use ImageInput with ``batch=False`` (the default):

    .. code-block:: python

        @api(input=ImageInput(), batch=False)
        def predict(self, image_array: imageio.core.utils.Array) -> str:
            results = self.artifacts.classifer.predict([image_array])
            return CLASS_NAMES[results[0]]

    Query with HTTP request::

        curl -i \\
          --header "Content-Type: image/jpeg" \\
          --request POST \\
          --data-binary @test.jpg \\
          localhost:5000/predict

    OR::

        curl -i \\
          -F image=@test.jpg \\
          localhost:5000/predict

    OR by an HTML form that sends multipart data:

    .. code-block:: html

        <form action="http://localhost:8000" method="POST"
              enctype="multipart/form-data">
            <input name="image" type="file">
            <input type="submit">
        </form>

    OR by python requests:

    .. code-block:: python

        import requests

        with open("test.jpg", "rb") as f:
            image_bytes = f.read()  # from file path

        files = {
            "image": ("test.jpg", image_bytes),
        }
        response = requests.post(your_url, files=files)

    .. code-block:: python

        import requests
        import PIL

        pil_image = PIL.Image.open('test.jpg')

        image_bytes = pil_image.tobytes()  # from PIL.Image

        files = {
            "image": ("test.jpg", image_bytes),
        }
        response = requests.post(your_url, files=files)

    Query with CLI command::

        bentoml run PyTorchFashionClassifier:latest predict --input-file test.jpg

    OR infer all images under a folder with ten images each batch::

        bentoml run PyTorchFashionClassifier:latest predict \\
          --input-file folder/*.jpg --max-batch-size 10
    """

    def __init__(
        self, accept_image_formats=None, pilmode="RGB", **base_kwargs,
    ):
        assert imageio, "`imageio` dependency can be imported"

        super().__init__(**base_kwargs)
        if 'input_names' in base_kwargs:
            raise TypeError(
                "ImageInput doesn't take input_names as parameters since bentoml 0.8."
                "Update your Service definition "
                "or use MultiImageInput instead."
            )

        self.pilmode = pilmode
        self.accept_image_formats = set(
            accept_image_formats or get_default_accept_image_formats()
        )

    @property
    def config(self):
        return {
            # Converting to list, google.protobuf.Struct does not work with tuple type
            "accept_image_formats": list(self.accept_image_formats),
            "pilmode": self.pilmode,
        }

    @property
    def request_schema(self):
        return {
            "image/*": {"schema": {"type": "string", "format": "binary"}},
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {
                        "image_file": {"type": "string", "format": "binary"}
                    },
                }
            },
        }

    @property
    def pip_dependencies(self):
        return ["imageio"]

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[BinaryIO]]
    ) -> ApiFuncArgs:
        img_list = []
        for task in tasks:
            if getattr(task.data, "name", None) and not check_file_extension(
                task.data.name, self.accept_image_formats
            ):
                task.discard(
                    http_status=400,
                    err_msg=f"Current service only accepts "
                    f"{self.accept_image_formats} formats",
                )
                continue
            try:
                img_array = imageio.imread(task.data, pilmode=self.pilmode)
                img_list.append(img_array)
            except ValueError as e:
                task.discard(http_status=400, err_msg=str(e))

        return (img_list,)



import os

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


XGBOOST_MODEL_EXTENSION = ".model"


class XgboostModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading Xgboost model

    Args:
        name (string): name of the artifact

    Raises:
        ImportError: xgboost package is required for using XgboostModelArtifact
        TypeError: invalid argument type, model being packed must be instance of
            xgboost.core.Booster

    Example usage:

    >>> import xgboost
    >>>
    >>> # prepare data
    >>> params = {... params}
    >>> dtrain = xgboost.DMatrix(...)
    >>>
    >>> # train model
    >>> model_to_save = xgboost.train(params=params, dtrain=dtrain)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.xgboost import XgboostModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts(XgboostModelArtifact('model'))
    >>> class XGBoostModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         result = self.artifacts.model.predict(df)
    >>>         return result
    >>>
    >>> svc = XGBoostModelService()
    >>> # Pack xgboost model
    >>> svc.pack('model', model_to_save)
    """

    def __init__(self, name):
        super(XgboostModelArtifact, self).__init__(name)
        self._model = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['xgboost'])

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + XGBOOST_MODEL_EXTENSION)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        try:
            import xgboost as xgb
        except ImportError:
            raise MissingDependencyException(
                "xgboost package is required to use XgboostModelArtifact"
            )

        if not isinstance(model, xgb.core.Booster):
            raise InvalidArgument(
                "Expect `model` argument to be a `xgboost.core.Booster` instance"
            )

        self._model = model
        return self

    def load(self, path):
        try:
            import xgboost as xgb
        except ImportError:
            raise MissingDependencyException(
                "xgboost package is required to use XgboostModelArtifact"
            )
        bst = xgb.Booster()
        bst.load_model(self._model_file_path(path))

        return self.pack(bst)

    def save(self, dst):
        return self._model.save_model(self._model_file_path(dst))

    def get(self):
        return self._model



import os
import re

from bentoml.service import BentoServiceArtifact


class TextFileArtifact(BentoServiceArtifact):
    """Abstraction for saving/loading string to/from text files

    Args:
        name (str): Name of the artifact
        file_extension (:obj:`str`, optional): The file extension used for the saved
            text file. Defaults to ".txt"
        encoding (str): The encoding will be used for saving/loading text. Defaults
            to "utf8"

    """

    def __init__(self, name, file_extension=".txt", encoding="utf-8"):
        super(TextFileArtifact, self).__init__(name)
        self._file_extension = file_extension
        self._encoding = encoding
        self._content = None

    def _text_file_path(self, base_path):
        return os.path.join(
            base_path,
            re.sub('[^-a-zA-Z0-9_.() ]+', '', self.name) + self._file_extension,
        )

    def load(self, path):
        with open(self._text_file_path(path), "rb") as f:
            content = f.read().decode(self._encoding)
        return self.pack(content)

    def pack(self, content, metadata=None):  # pylint:disable=arguments-renamed
        self._content = content
        return self

    def get(self):
        return self._content

    def save(self, dst):
        with open(self._text_file_path(dst), "wb") as f:
            f.write(self._content.encode(self._encoding))
